import os
import pathlib
import shutil
import tempfile
from typing import Any

from ruyi.ruyipkg.state import BoundInstallationStateStore

from ..cli.user_input import ask_for_yesno_confirmation
from ..config import GlobalConfig
from ..i18n import _
from ..telemetry.scope import TelemetryScope
from .atom import Atom
from .distfile import Distfile
from .host import RuyiHost
from .pkg_manifest import BoundPackageManifest
from .composite_repo import CompositeRepo
from .unpack import ensure_unpack_cmd_for_method


def is_root_likely_populated(root: str) -> bool:
    try:
        return any(os.scandir(root))
    except FileNotFoundError:
        return False


def do_extract_atoms(
    cfg: GlobalConfig,
    mr: CompositeRepo,
    atom_strs: set[str],
    *,
    canonicalized_host: str | RuyiHost,
    dest_dir: os.PathLike[Any] | None,  # None for CWD
    extract_without_subdir: bool,
    fetch_only: bool,
) -> int:
    logger = cfg.logger
    logger.D(f"about to extract for host {canonicalized_host}: {atom_strs}")

    mr = cfg.repo

    for a_str in atom_strs:
        a = Atom.parse(a_str)
        pm = a.match_in_repo(mr, cfg.include_prereleases)
        if pm is None:
            logger.F(
                _("atom {atom} matches no package in the repository").format(
                    atom=a_str,
                )
            )
            return 1

        sv = pm.service_level
        if sv.has_known_issues:
            logger.W(_("package has known issue(s)"))
            for s in sv.render_known_issues(pm.repo.messages, cfg.lang_code):
                logger.I(s)

        ret = _do_extract_pkg(
            cfg,
            pm,
            canonicalized_host=canonicalized_host,
            fetch_only=fetch_only,
            dest_dir=dest_dir,
            extract_without_subdir=extract_without_subdir,
        )
        if ret != 0:
            return ret

    return 0


def _do_extract_pkg(
    cfg: GlobalConfig,
    pm: BoundPackageManifest,
    *,
    canonicalized_host: str | RuyiHost,
    dest_dir: os.PathLike[Any] | None,  # None for CWD
    extract_without_subdir: bool,
    fetch_only: bool,
) -> int:
    logger = cfg.logger

    pkg_name = pm.name_for_installation

    if not extract_without_subdir:
        # extract into a subdirectory named <pkg_name>-<version>
        subdir_name = pm.name_for_installation
        if dest_dir is None:
            dest_dir = pathlib.Path(subdir_name)
        else:
            dest_dir = pathlib.Path(dest_dir) / subdir_name

    logger.D(f"about to extract {pm} to {dest_dir}")

    # Make sure destination directory exists
    if dest_dir is not None:
        dest_dir = pathlib.Path(dest_dir)
        dest_dir.mkdir(parents=True, exist_ok=True)

    bm = pm.binary_metadata
    sm = pm.source_metadata
    if bm is None and sm is None:
        logger.F(
            _("don't know how to extract package [green]{pkg}[/]").format(
                pkg=pkg_name,
            )
        )
        return 2

    if bm is not None and sm is not None:
        logger.F(
            _(
                "cannot handle package [green]{pkg}[/]: package is both binary and source"
            ).format(
                pkg=pkg_name,
            )
        )
        return 2

    distfiles_for_host: list[str] | None = None
    if bm is not None:
        distfiles_for_host = bm.get_distfile_names_for_host(canonicalized_host)
    elif sm is not None:
        distfiles_for_host = sm.get_distfile_names_for_host(canonicalized_host)

    if not distfiles_for_host:
        logger.F(
            _("package [green]{pkg}[/] declares no distfile for host {host}").format(
                pkg=pkg_name,
                host=canonicalized_host,
            )
        )
        return 2

    dfs = pm.distfiles

    for df_name in distfiles_for_host:
        df_decl = dfs[df_name]
        ensure_unpack_cmd_for_method(logger, df_decl.unpack_method)
        df = Distfile(df_decl, pm.repo)
        df.ensure(logger)

        if fetch_only:
            logger.D("skipping extraction because [yellow]--fetch-only[/] is given")
            continue

        logger.I(
            _("extracting [green]{distfile}[/] for package [green]{pkg}[/]").format(
                distfile=df_name,
                pkg=pkg_name,
            )
        )
        # unpack into destination
        df.unpack(dest_dir, logger)

    if not fetch_only:
        # None is the unpack helper's CWD sentinel; report it as a real path.
        reported_dest_dir = pathlib.Path(".") if dest_dir is None else dest_dir
        logger.I(
            _("package [green]{pkg}[/] has been extracted to {dest_dir}").format(
                pkg=pkg_name,
                dest_dir=reported_dest_dir,
            )
        )

    return 0


def do_install_atoms(
    config: GlobalConfig,
    mr: CompositeRepo,
    atom_strs: set[str],
    *,
    canonicalized_host: str | RuyiHost,
    fetch_only: bool,
    reinstall: bool,
) -> int:
    logger = config.logger
    logger.D(f"about to install for host {canonicalized_host}: {atom_strs}")

    for a_str in atom_strs:
        a = Atom.parse(a_str)
        pm = a.match_in_repo(mr, config.include_prereleases)
        if pm is None:
            logger.F(
                _("atom {atom} matches no package in the repository").format(atom=a_str)
            )
            return 1
        pkg_name = pm.name_for_installation

        sv = pm.service_level
        if sv.has_known_issues:
            logger.W(_("package has known issue(s)"))
            for s in sv.render_known_issues(pm.repo.messages, config.lang_code):
                logger.I(s)

        config.telemetry.record(
            TelemetryScope(mr.repo_id),
            "repo:package-install-v1",
            atom=a_str,
            host=canonicalized_host,
            pkg_category=pm.category,
            pkg_kinds=pm.kind,
            pkg_name=pm.name,
            pkg_version=pm.ver,
        )

        if pm.binary_metadata is not None:
            ret = _do_install_binary_pkg(
                config,
                mr,
                pm,
                canonicalized_host,
                fetch_only,
                reinstall,
            )
            if ret != 0:
                return ret
            continue

        if pm.blob_metadata is not None:
            ret = _do_install_blob_pkg(config, mr, pm, fetch_only, reinstall)
            if ret != 0:
                return ret
            continue

        # the user may be trying to fetch a source-only package with `ruyi install --fetch-only`,
        # so try that too for better UX
        if fetch_only and pm.source_metadata is not None:
            ret = _do_extract_pkg(
                config,
                pm,
                canonicalized_host=canonicalized_host,
                dest_dir=None,  # unused in this case
                extract_without_subdir=False,  # unused in this case
                fetch_only=fetch_only,
            )
            if ret != 0:
                return ret
            continue

        logger.F(
            _("don't know how to handle non-binary package [green]{pkg}[/]").format(
                pkg=pkg_name,
            )
        )
        return 2

    return 0


def _do_install_binary_pkg(
    config: GlobalConfig,
    mr: CompositeRepo,
    pm: BoundPackageManifest,
    canonicalized_host: str | RuyiHost,
    fetch_only: bool,
    reinstall: bool,
) -> int:
    logger = config.logger
    bm = pm.binary_metadata
    assert bm is not None

    pkg_name = pm.name_for_installation
    install_root = config.global_binary_install_root(str(canonicalized_host), pkg_name)

    rgs = config.ruyipkg_global_state
    is_installed = rgs.is_package_installed(
        pm.repo_id,
        pm.category,
        pm.name,
        pm.ver,
        str(canonicalized_host),
    )

    # Fallback to directory check if not tracked in state
    if not is_installed and is_root_likely_populated(install_root):
        is_installed = True

    if is_installed:
        if not reinstall:
            logger.I(
                _("skipping already installed package [green]{pkg}[/]").format(
                    pkg=pkg_name,
                )
            )
            return 0

        logger.W(
            _(
                "package [green]{pkg}[/] seems already installed; purging and re-installing due to [yellow]--reinstall[/]"
            ).format(pkg=pkg_name)
        )
        # Remove from state tracking before purging
        rgs.remove_installation(
            pm.repo_id,
            pm.category,
            pm.name,
            pm.ver,
            str(canonicalized_host),
        )
        shutil.rmtree(install_root)

    ir_parent = pathlib.Path(install_root).resolve().parent
    ir_parent.mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(prefix=".ruyi-tmp", dir=ir_parent) as tmp_root:
        ret = _do_install_binary_pkg_to(
            config,
            mr,
            pm,
            canonicalized_host,
            fetch_only,
            tmp_root,
        )
        if ret != 0:
            return ret
        os.rename(tmp_root, install_root)

    if not fetch_only:
        rgs.record_installation(
            repo_id=pm.repo_id,
            category=pm.category,
            name=pm.name,
            version=pm.ver,
            host=str(canonicalized_host),
            install_path=install_root,
        )

    repo_tag = f" [dim]\\[{pm.repo_id}][/]" if len(config.repo_entries) > 1 else ""
    logger.I(
        _(
            "package [green]{pkg}[/]{repo_tag} installed to [yellow]{install_root}[/]"
        ).format(
            pkg=pkg_name,
            repo_tag=repo_tag,
            install_root=install_root,
        )
    )
    return 0


def _do_install_binary_pkg_to(
    config: GlobalConfig,
    mr: CompositeRepo,
    pm: BoundPackageManifest,
    canonicalized_host: str | RuyiHost,
    fetch_only: bool,
    install_root: str,
) -> int:
    logger = config.logger
    bm = pm.binary_metadata
    assert bm is not None

    dfs = pm.distfiles

    pkg_name = pm.name_for_installation
    distfiles_for_host = bm.get_distfile_names_for_host(str(canonicalized_host))
    if not distfiles_for_host:
        logger.F(
            _("package [green]{pkg}[/] declares no binary for host {host}").format(
                pkg=pkg_name,
                host=canonicalized_host,
            )
        )
        return 2

    for df_name in distfiles_for_host:
        df_decl = dfs[df_name]
        ensure_unpack_cmd_for_method(logger, df_decl.unpack_method)
        df = Distfile(df_decl, pm.repo)
        df.ensure(logger)

        if fetch_only:
            logger.D(
                _("skipping installation because [yellow]--fetch-only[/] is given")
            )
            continue

        logger.I(
            _("extracting [green]{distfile}[/] for package [green]{pkg}[/]").format(
                distfile=df_name,
                pkg=pkg_name,
            )
        )
        df.unpack(install_root, logger)

    return 0


def _do_install_blob_pkg(
    config: GlobalConfig,
    mr: CompositeRepo,
    pm: BoundPackageManifest,
    fetch_only: bool,
    reinstall: bool,
) -> int:
    logger = config.logger
    bm = pm.blob_metadata
    assert bm is not None

    pkg_name = pm.name_for_installation
    install_root = config.global_blob_install_root(pkg_name)

    rgs = config.ruyipkg_global_state
    is_installed = rgs.is_package_installed(
        pm.repo_id,
        pm.category,
        pm.name,
        pm.ver,
        "",  # host is "" for blob packages
    )

    # Fallback to directory check if not tracked in state
    if not is_installed and is_root_likely_populated(install_root):
        is_installed = True

    if is_installed:
        if not reinstall:
            logger.I(
                _("skipping already installed package [green]{pkg}[/]").format(
                    pkg=pkg_name,
                )
            )
            return 0

        logger.W(
            _(
                "package [green]{pkg}[/] seems already installed; purging and re-installing due to [yellow]--reinstall[/]"
            ).format(pkg=pkg_name)
        )
        # Remove from state tracking before purging
        rgs.remove_installation(
            pm.repo_id,
            pm.category,
            pm.name,
            pm.ver,
            "",
        )
        shutil.rmtree(install_root)

    ir_parent = pathlib.Path(install_root).resolve().parent
    ir_parent.mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(prefix=".ruyi-tmp", dir=ir_parent) as tmp_root:
        ret = _do_install_blob_pkg_to(
            config,
            mr,
            pm,
            fetch_only,
            tmp_root,
        )
        if ret != 0:
            return ret
        os.rename(tmp_root, install_root)

    if not fetch_only:
        rgs.record_installation(
            repo_id=pm.repo_id,
            category=pm.category,
            name=pm.name,
            version=pm.ver,
            host="",  # Empty for blob packages
            install_path=install_root,
        )

    logger.I(
        _("package [green]{pkg}[/] installed to [yellow]{install_root}[/]").format(
            pkg=pkg_name,
            install_root=install_root,
        )
    )
    return 0


def _do_install_blob_pkg_to(
    config: GlobalConfig,
    mr: CompositeRepo,
    pm: BoundPackageManifest,
    fetch_only: bool,
    install_root: str,
) -> int:
    logger = config.logger
    bm = pm.blob_metadata
    assert bm is not None

    pkg_name = pm.name_for_installation
    dfs = pm.distfiles
    distfile_names = bm.get_distfile_names()
    if not distfile_names:
        logger.F(
            _("package [green]{pkg}[/] declares no blob distfile").format(pkg=pkg_name)
        )
        return 2

    for df_name in distfile_names:
        df_decl = dfs[df_name]
        ensure_unpack_cmd_for_method(logger, df_decl.unpack_method)
        df = Distfile(df_decl, pm.repo)
        df.ensure(logger)

        if fetch_only:
            logger.D(
                _("skipping installation because [yellow]--fetch-only[/] is given")
            )
            continue

        logger.I(
            _("extracting [green]{distfile}[/] for package [green]{pkg}[/]").format(
                distfile=df_name,
                pkg=pkg_name,
            )
        )
        df.unpack_or_symlink(install_root, logger)

    return 0


def do_uninstall_atoms(
    config: GlobalConfig,
    mr: CompositeRepo,
    atom_strs: set[str],
    *,
    canonicalized_host: str | RuyiHost,
    assume_yes: bool,
) -> int:
    logger = config.logger
    logger.D(f"about to uninstall for host {canonicalized_host}: {atom_strs}")

    bis = BoundInstallationStateStore(config.ruyipkg_global_state, mr)

    pms_to_uninstall: list[tuple[str, BoundPackageManifest]] = []
    for a_str in atom_strs:
        a = Atom.parse(a_str)
        pm = a.match_in_repo(bis, config.include_prereleases)
        if pm is None:
            logger.F(
                _("atom [yellow]{atom}[/] is non-existent or not installed").format(
                    atom=a_str,
                )
            )
            return 1
        pms_to_uninstall.append((a_str, pm))

    if not pms_to_uninstall:
        logger.I(_("no packages to uninstall"))
        return 0

    logger.I(_("the following packages will be uninstalled:"))
    for _unused, pm in pms_to_uninstall:
        logger.I(
            _("  - [green]{category}/{name}[/] ({version})").format(
                category=pm.category,
                name=pm.name,
                version=pm.ver,
            )
        )

    if not assume_yes:
        if not ask_for_yesno_confirmation(logger, _("Proceed?"), default=False):
            logger.I(_("uninstallation aborted"))
            return 0

    for a_str, pm in pms_to_uninstall:
        pkg_name = pm.name_for_installation

        config.telemetry.record(
            TelemetryScope(mr.repo_id),
            "repo:package-uninstall-v1",
            atom=a_str,
            host=canonicalized_host,
            pkg_category=pm.category,
            pkg_kinds=pm.kind,
            pkg_name=pm.name,
            pkg_version=pm.ver,
        )

        if pm.binary_metadata is not None:
            ret = _do_uninstall_binary_pkg(
                config,
                pm,
                canonicalized_host,
            )
            if ret != 0:
                return ret
            continue

        if pm.blob_metadata is not None:
            ret = _do_uninstall_blob_pkg(config, pm)
            if ret != 0:
                return ret
            continue

        logger.F(
            _("don't know how to handle non-binary package [green]{pkg}[/]").format(
                pkg=pkg_name,
            )
        )
        return 2

    return 0


def _do_uninstall_binary_pkg(
    config: GlobalConfig,
    pm: BoundPackageManifest,
    canonicalized_host: str | RuyiHost,
) -> int:
    logger = config.logger
    bm = pm.binary_metadata
    assert bm is not None

    pkg_name = pm.name_for_installation
    install_root = config.global_binary_install_root(str(canonicalized_host), pkg_name)

    rgs = config.ruyipkg_global_state
    is_installed = rgs.is_package_installed(
        pm.repo_id,
        pm.category,
        pm.name,
        pm.ver,
        str(canonicalized_host),
    )

    # Check directory existence if the PM state says the package is not installed
    if not is_installed:
        if not os.path.exists(install_root):
            logger.I(
                _("skipping not-installed package [green]{pkg}[/]").format(
                    pkg=pkg_name,
                )
            )
            return 0

        # There may be potentially user-generated data in the directory,
        # let's be safe and fail the process.
        logger.F(
            _(
                "package [green]{pkg}[/] is not tracked as installed, but its directory [yellow]{install_root}[/] exists."
            ).format(pkg=pkg_name, install_root=install_root)
        )
        logger.I(_("Please remove it manually if you are sure it's safe to do so."))
        logger.I(
            _(
                "If you believe this is a bug, please file an issue at [yellow]https://github.com/ruyisdk/ruyi/issues[/]."
            )
        )
        return 1

    logger.I(_("uninstalling package [green]{pkg}[/]").format(pkg=pkg_name))
    if is_installed:
        rgs.remove_installation(
            pm.repo_id,
            pm.category,
            pm.name,
            pm.ver,
            str(canonicalized_host),
        )

    if os.path.exists(install_root):
        shutil.rmtree(install_root)

    logger.I(_("package [green]{pkg}[/] uninstalled").format(pkg=pkg_name))
    return 0


def _do_uninstall_blob_pkg(
    config: GlobalConfig,
    pm: BoundPackageManifest,
) -> int:
    logger = config.logger
    bm = pm.blob_metadata
    assert bm is not None

    pkg_name = pm.name_for_installation
    install_root = config.global_blob_install_root(pkg_name)

    rgs = config.ruyipkg_global_state
    is_installed = rgs.is_package_installed(
        pm.repo_id,
        pm.category,
        pm.name,
        pm.ver,
        "",  # host is "" for blob packages
    )

    # Check directory existence if the PM state says the package is not installed
    if not is_installed:
        if not os.path.exists(install_root):
            logger.I(
                _("skipping not-installed package [green]{pkg}[/]").format(
                    pkg=pkg_name,
                )
            )
            return 0

        # There may be potentially user-generated data in the directory,
        # let's be safe and fail the process.
        logger.F(
            _(
                "package [green]{pkg}[/] is not tracked as installed, but its directory [yellow]{install_root}[/] exists."
            ).format(pkg=pkg_name, install_root=install_root)
        )
        logger.I(_("Please remove it manually if you are sure it's safe to do so."))
        logger.I(
            _(
                "If you believe this is a bug, please file an issue at [yellow]https://github.com/ruyisdk/ruyi/issues[/]."
            )
        )
        return 1

    logger.I(_("uninstalling package [green]{pkg}[/]").format(pkg=pkg_name))
    if is_installed:
        rgs.remove_installation(
            pm.repo_id,
            pm.category,
            pm.name,
            pm.ver,
            "",
        )

    if os.path.exists(install_root):
        shutil.rmtree(install_root)

    logger.I(_("package [green]{pkg}[/] uninstalled").format(pkg=pkg_name))
    return 0



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import AdminCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class AdminRunPluginCommand(
    AdminCommand,
    cmd="run-plugin-cmd",
    help=_("Run a plugin-defined command"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "cmd_name",
            type=str,
            metavar="COMMAND-NAME",
            help=_("Command name"),
        )
        p.add_argument(
            "cmd_args",
            type=str,
            nargs="*",
            metavar="COMMAND-ARG",
            help=_("Arguments to pass to the plugin command"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        cmd_name = args.cmd_name
        cmd_args = args.cmd_args

        return cfg.repo.run_plugin_cmd(cmd_name, cmd_args)



from typing import Literal, TypeAlias, TypeGuard

TelemetryScopeConfig: TypeAlias = Literal["pm"] | Literal["repo"]


def is_telemetry_scope_config(x: object) -> TypeGuard[TelemetryScopeConfig]:
    if not isinstance(x, str):
        return False
    match x:
        case "pm" | "repo":
            return True
        case _:
            return False


class TelemetryScope:
    def __init__(self, repo_name: str | None) -> None:
        self._repo_name = repo_name

    def __repr__(self) -> str:
        return f"TelemetryScope(repo_name={self._repo_name})"

    def __str__(self) -> str:
        if self._repo_name:
            return f"repo:{self._repo_name}"
        return "pm"

    def __hash__(self) -> int:
        # behave like the inner field
        return hash(self._repo_name)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, TelemetryScope):
            return False
        return self._repo_name == value._repo_name

    @property
    def repo_name(self) -> str | None:
        return self._repo_name

    @property
    def is_pm(self) -> bool:
        return self._repo_name is None



import locale

from typing import Iterable, NamedTuple


class LangAndRegion(NamedTuple):
    raw: str
    lang: str
    region: str | None


def lang_code_to_lang_region(lang_code: str, guess_region: bool) -> LangAndRegion:
    if not guess_region and "_" not in lang_code:
        return LangAndRegion(lang_code, lang_code, None)

    lang_region_str = locale.normalize(lang_code).split(".")[0]
    parts = lang_region_str.split("_", 2)
    if len(parts) == 1:
        return LangAndRegion(lang_code, lang_region_str, None)
    return LangAndRegion(lang_code, parts[0], parts[1])


def match_lang_code(
    req: str,
    avail: Iterable[str],
) -> str:
    """Returns a proper available language code based on a list of available
    language codes, and a request."""

    if not isinstance(avail, set) or not isinstance(avail, frozenset):
        avail = set(avail)

    # return the only one choice if this is the case
    if len(avail) == 1:
        return next(iter(avail))

    # try exact match
    if req in avail:
        return req

    return _match_lang_code_slowpath(
        lang_code_to_lang_region(req, True),
        [lang_code_to_lang_region(x, False) for x in avail],
    )


def _match_lang_code_slowpath(
    req: LangAndRegion,
    avail: list[LangAndRegion],
) -> str:
    # pick one with the requested region
    if req.region is not None:
        for x in avail:
            if x.region == req.region:
                return x.raw

    # if no match, pick one with the requested language
    for x in avail:
        if x.lang == req.lang:
            return x.raw

    # neither matches, fallback to (en_US, en, en_*, zh_CN, zh, zh_*)
    # in that order
    fallback_en = {x.region: x.raw for x in avail if x.lang == "en"}
    if fallback_en:
        if "US" in fallback_en:
            return fallback_en["US"]
        if None in fallback_en:
            return fallback_en[None]
        return fallback_en[sorted(x for x in fallback_en.keys() if x is not None)[0]]

    fallback_zh = {x.region: x.raw for x in avail if x.lang == "zh"}
    if fallback_zh:
        if "CN" in fallback_zh:
            return fallback_zh["CN"]
        if None in fallback_zh:
            return fallback_zh[None]
        return fallback_zh[sorted(x for x in fallback_zh.keys() if x is not None)[0]]

    # neither en nor zh is available (which is highly unlikely at present)
    # pick the first available one as a last resort
    # sort the list before picking for determinism
    return sorted(x.raw for x in avail)[0]



"""Utilities for parsing mount information from /proc/self/mounts."""

import pathlib
import re
from typing import NamedTuple


class MountInfo(NamedTuple):
    source: str
    target: str
    fstype: str
    options: list[str]

    @property
    def source_path(self) -> pathlib.Path:
        return pathlib.Path(self.source)

    @property
    def source_is_blkdev(self) -> bool:
        return self.source_path.is_block_device()


def parse_mounts(contents: str | None = None) -> list[MountInfo]:
    if contents is None:
        try:
            with open("/proc/self/mounts", "r", encoding="utf-8") as f:
                contents = f.read()
        except OSError:
            return []

    mounts: list[MountInfo] = []
    for line in contents.splitlines():
        parts = line.split()
        if len(parts) < 4:
            continue
        source, target, fstype, opts = parts[:4]
        options = opts.split(",")
        source = _unescape_octals(source)
        target = _unescape_octals(target)
        mounts.append(MountInfo(source, target, fstype, options))
    return mounts


def _unescape_octals(s: str) -> str:
    return re.sub(r"\\([0-3][0-7]{2})", lambda m: chr(int(m.group(1), 8)), s)



from contextlib import AbstractContextManager
from types import TracebackType
from typing import Iterable

import tomlkit
from tomlkit.container import Container
from tomlkit.items import Array, Comment, InlineTable, Item, Table, Trivia, Whitespace


class NoneValue(Exception):
    """Used to indicate that a None value is to be dumped in TOML. Because TOML
    does not support None natively, this means special handling is needed."""

    def __str__(self) -> str:
        return "NoneValue()"

    def __repr__(self) -> str:
        return "NoneValue()"


def with_indent(item: Item, spaces: int = 2) -> Item:
    item.indent(spaces)
    return item


def inline_table_with_spaces() -> "InlineTableWithSpaces":
    return InlineTableWithSpaces(Container(), Trivia(), new=True)


class InlineTableWithSpaces(InlineTable, AbstractContextManager[InlineTable]):
    def __init__(
        self,
        value: Container,
        trivia: Trivia,
        new: bool = False,
    ) -> None:
        super().__init__(value, trivia, new)

    def __enter__(self) -> InlineTable:
        self.add(tomlkit.ws(" "))
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        self.add(tomlkit.ws(" "))
        return None


def _into_item(x: Item | str) -> Item:
    if isinstance(x, Item):
        return x
    return tomlkit.string(x)


def str_array(
    args: Iterable[Item | str],
    *,
    multiline: bool = False,
    indent: int = 2,
) -> Array:
    items = [_into_item(i).indent(indent) for i in args]
    return Array(items, Trivia(), multiline=multiline)


def sorted_table(x: dict[str, str]) -> Table:
    y = tomlkit.table()
    for k in sorted(x.keys()):
        y.add(k, x[k])
    return y


def extract_header_comments(
    doc: Container,
) -> list[str]:
    comments: list[str] = []

    # ignore leading whitespaces
    is_skipping_leading_ws = True
    for _key, item in doc.body:
        if isinstance(item, Whitespace):
            if is_skipping_leading_ws:
                continue
            # this is part of the header comments
            comments.append(item.as_string())
        elif isinstance(item, Comment):
            is_skipping_leading_ws = False
            comments.append(item.as_string())
        else:
            # we reached the first non-comment item
            break
    return comments


def extract_footer_comments(
    doc: Container,
) -> list[str]:
    comments: list[str] = []

    # ignore trailing whitespaces
    is_skipping_trailing_ws = True
    for _key, item in reversed(doc.body):
        if isinstance(item, Whitespace):
            if is_skipping_trailing_ws:
                continue
            # this is part of the footer comments
            comments.append(item.as_string())
        elif isinstance(item, Comment):
            is_skipping_trailing_ws = False
            comments.append(item.as_string())
        else:
            # we reached the first non-comment item
            break

    # if the footer comment was preceded by a table, then the comment would be
    # nested inside the table and invisible in top-level doc.body, so we would
    # have to check the last item as well
    if not comments:
        last_elem = doc.body[-1][1].value
        if isinstance(last_elem, Container):
            return extract_footer_comments(last_elem)

    return list(reversed(comments))



from contextlib import AbstractContextManager
import pathlib
from typing import Final, Sequence, TYPE_CHECKING, cast

if TYPE_CHECKING:
    from types import TracebackType
    from typing_extensions import Self

import tomlkit
from tomlkit.items import AoT, Table

from .errors import MalformedConfigFileError, ProtectedGlobalConfigError
from .schema import (
    SECTION_INSTALLATION,
    SECTION_REPOS,
    KEY_INSTALLATION_EXTERNALLY_MANAGED,
    KEY_REPOS_ID,
    ensure_valid_config_kv,
    parse_config_key,
    validate_section,
)

if TYPE_CHECKING:
    from . import GlobalConfig


GLOBAL_ONLY_CONFIG_KEYS: Final[set[tuple[str, str]]] = {
    (SECTION_INSTALLATION, KEY_INSTALLATION_EXTERNALLY_MANAGED),
}
"""Settings that can only be set in global-scope config files.

Changes should be reflected in ``GlobalConfig._apply_config`` too."""


def _is_config_key_global_only(key: str | Sequence[str]) -> bool:
    parsed_key = parse_config_key(key)
    if len(parsed_key) != 2:
        return False
    section, leaf = parsed_key
    return (section, leaf) in GLOBAL_ONLY_CONFIG_KEYS


class ConfigEditor(AbstractContextManager["ConfigEditor"]):
    def __init__(self, path: pathlib.Path) -> None:
        self._path = path
        self._touched = False
        try:
            with open(path) as fp:
                self._content = tomlkit.load(fp)
            if not isinstance(self._content, tomlkit.TOMLDocument):
                raise MalformedConfigFileError(path)
        except FileNotFoundError:
            self._content = tomlkit.document()

        self._stage = cast(tomlkit.TOMLDocument, self._content.copy())  # type: ignore[redundant-cast,unused-ignore]

    @classmethod
    def work_on_user_local_config(cls, gc: "GlobalConfig") -> "Self":
        return cls(gc.local_user_config_file)

    def __enter__(self) -> "Self":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: "TracebackType | None",
    ) -> bool | None:
        self._commit()
        return None

    def _commit(self) -> None:
        if not self._touched:
            return

        self._path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._path, "w", encoding="utf-8") as fp:
            tomlkit.dump(self._content, fp)

    def stage(self) -> None:
        self._content = self._stage
        self._touched = True
        self._stage = cast(tomlkit.TOMLDocument, self._content.copy())  # type: ignore[redundant-cast,unused-ignore]

    def set_value(self, key: str | Sequence[str], val: object | None) -> None:
        parsed_key = parse_config_key(key)
        ensure_valid_config_kv(parsed_key, check_val=True, val=val)

        # Gate protected settings: user-local config is not allowed to override
        # global-only settings.
        if _is_config_key_global_only(parsed_key):
            # This mechanism is only meant for modifying user-local configs,
            # because global configs are assumed to be maintained by packagers
            # and/or sysadmins, who are expected to write the config file by
            # hand.
            raise ProtectedGlobalConfigError(parsed_key)

        section, sel = parsed_key[0], parsed_key[1:]
        if section in self._stage:
            existing_section = self._stage[section]
            if not isinstance(existing_section, Table):
                raise MalformedConfigFileError(self._path)
            existing_section.update({sel[0]: val})
        else:
            # append a section with its sole key set to val
            new_section = tomlkit.table()
            new_section.append(sel[0], val)
            self._stage.append(section, new_section)

    def unset_value(self, key: str | Sequence[str]) -> None:
        parsed_key = parse_config_key(key)
        ensure_valid_config_kv(parsed_key, check_val=False)

        section, sel = parsed_key[0], parsed_key[1:]
        if existing_section := self._stage.get(section):
            if not isinstance(existing_section, Table):
                raise MalformedConfigFileError(self._path)
            if sel[0] in existing_section:
                existing_section.pop(sel[0])

    def remove_section(self, section: str) -> None:
        validate_section(section)
        if section in self._stage:
            self._stage.pop(section)

    def add_repos_entry(self, entry: dict[str, object]) -> None:
        """Append a ``[[repos]]`` table-array entry to the staged config."""
        tbl = tomlkit.table()
        for k, v in entry.items():
            tbl.append(k, v)

        if SECTION_REPOS in self._stage:
            existing = self._stage[SECTION_REPOS]
            if isinstance(existing, AoT):
                existing.append(tbl)
            else:
                raise MalformedConfigFileError(self._path)
        else:
            aot = tomlkit.aot()
            aot.append(tbl)
            self._stage.append(SECTION_REPOS, aot)

    def remove_repos_entry(self, repo_id: str) -> bool:
        """Remove the ``[[repos]]`` entry with the given *repo_id*.

        Returns True if an entry was removed, False if not found."""
        if SECTION_REPOS not in self._stage:
            return False
        existing = self._stage[SECTION_REPOS]
        if not isinstance(existing, AoT):
            raise MalformedConfigFileError(self._path)

        for i, tbl in enumerate(existing):
            if tbl.get(KEY_REPOS_ID) == repo_id:
                del existing[i]
                if len(existing) == 0:
                    self._stage.pop(SECTION_REPOS)
                return True
        return False

    def update_repos_entry(self, repo_id: str, updates: dict[str, object]) -> bool:
        """Update fields of the ``[[repos]]`` entry with the given *repo_id*.

        Returns True if the entry was found and updated, False otherwise."""
        if SECTION_REPOS not in self._stage:
            return False
        existing = self._stage[SECTION_REPOS]
        if not isinstance(existing, AoT):
            raise MalformedConfigFileError(self._path)

        for tbl in existing:
            if tbl.get(KEY_REPOS_ID) == repo_id:
                for k, v in updates.items():
                    tbl[k] = v
                return True
        return False



import argparse
from typing import TYPE_CHECKING

from ..i18n import _
from .cmd import RootCommand

if TYPE_CHECKING:
    from .completion import ArgumentParser
    from .. import config


# Config management commands
class ConfigCommand(
    RootCommand,
    cmd="config",
    has_subcommands=True,
    help=_("Manage Ruyi's config options"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        pass


class ConfigGetCommand(
    ConfigCommand,
    cmd="get",
    help=_("Query the value of a Ruyi config option"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "key",
            type=str,
            help=_("The Ruyi config option to query"),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.errors import InvalidConfigKeyError
        from ..config.schema import encode_value
        from ..utils.toml import NoneValue

        key: str = args.key

        try:
            val = cfg.get_by_key(key)
        except InvalidConfigKeyError:
            return 1

        try:
            encoded_val = encode_value(val)
        except NoneValue:
            return 1

        cfg.logger.stdout(encoded_val)
        return 0


class ConfigSetCommand(
    ConfigCommand,
    cmd="set",
    help=_("Set the value of a Ruyi config option"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "key",
            type=str,
            help=_("The Ruyi config option to set"),
        )
        p.add_argument(
            "value",
            type=str,
            help=_("The value to set the option to"),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor
        from ..config.errors import ProtectedGlobalConfigError
        from ..config.schema import decode_value

        key: str = args.key
        val: str = args.value

        pyval = decode_value(key, val)
        with ConfigEditor.work_on_user_local_config(cfg) as ed:
            try:
                ed.set_value(key, pyval)
            except ProtectedGlobalConfigError:
                cfg.logger.F(
                    _(
                        "the config [yellow]{key}[/] is protected and not meant to be overridden by users"
                    ).format(key=key)
                )
                return 2

            ed.stage()

        return 0


class ConfigUnsetCommand(
    ConfigCommand,
    cmd="unset",
    help=_("Unset a Ruyi config option"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "key",
            type=str,
            help=_("The Ruyi config option to unset"),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor

        key: str = args.key

        with ConfigEditor.work_on_user_local_config(cfg) as ed:
            ed.unset_value(key)
            ed.stage()

        return 0


class ConfigRemoveSectionCommand(
    ConfigCommand,
    cmd="remove-section",
    help=_("Remove a section from the Ruyi config"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "section",
            type=str,
            help=_("The section to remove"),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor

        section: str = args.section

        with ConfigEditor.work_on_user_local_config(cfg) as ed:
            ed.remove_section(section)
            ed.stage()

        return 0



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class EntityCommand(
    RootCommand,
    cmd="entity",
    has_subcommands=True,
    is_experimental=True,
    help=_("Interact with entities defined in the repositories"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass


class EntityDescribeCommand(
    EntityCommand,
    cmd="describe",
    help=_("Describe an entity"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "ref",
            help=_(
                "Reference to the entity to describe in the form of '<type>:<name>'"
            ),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        logger = cfg.logger
        ref = args.ref

        entity_store = cfg.repo.entity_store
        entity = entity_store.get_entity_by_ref(ref)
        if entity is None:
            logger.F(_("entity [yellow]{ref}[/] not found").format(ref=ref))
            return 1

        logger.stdout(
            _("Entity [bold]{entity}[/] ([green]{display_name}[/])\n").format(
                entity=str(entity),
                display_name=entity.display_name,
            )
        )

        fwd_refs = entity.related_refs
        if fwd_refs:
            logger.stdout(_("  Direct forward relationships:"))
            for ref in sorted(fwd_refs):
                logger.stdout(f"    - [yellow]{ref}[/]")
        else:
            logger.stdout(_("  Direct forward relationships: [gray]none[/]"))

        rev_refs = entity.reverse_refs
        if rev_refs:
            logger.stdout(_("  Direct reverse relationships:"))
            for ref in sorted(rev_refs):
                logger.stdout(f"    - [yellow]{ref}[/]")
        else:
            logger.stdout(_("  Direct reverse relationships: [gray]none[/]"))

        logger.stdout(_("  All indirectly related entities:"))
        for e in entity_store.traverse_related_entities(
            entity,
            transitive=True,
            no_direct_refs=True,
            forward_refs=True,
            reverse_refs=True,
        ):
            logger.stdout(f"    - [yellow]{e}[/]")

        # TODO: render type-specific data

        return 0


class EntityListCommand(
    EntityCommand,
    cmd="list",
    help=_("List entities"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "-t",
            "--entity-type",
            action="append",
            nargs=1,
            dest="entity_type",
            help=_(
                "List entities of this type. Can be passed multiple times to list multiple types."
            ),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ..utils.porcelain import PorcelainOutput

        entity_types_in: list[list[str]] | None = args.entity_type
        entity_types: list[str] | None = None
        if entity_types_in is not None:
            entity_types = [x[0] for x in entity_types_in]

        logger = cfg.logger
        entity_store = cfg.repo.entity_store

        # Check if porcelain output is requested
        if cfg.is_porcelain:
            with PorcelainOutput() as po:
                for e in entity_store.iter_entities(entity_types):
                    po.emit(e.to_porcelain())
            return 0

        # Human-readable output
        for e in entity_store.iter_entities(entity_types):
            logger.stdout(f"'{str(e)}':")
            logger.stdout(f"  display name: {e.display_name}")
            logger.stdout(f"  data: {e.data}")
            logger.stdout(f"  forward_refs: {e.related_refs}")
            logger.stdout(f"  reverse_refs: {e.reverse_refs}")
        return 0



import datetime
from typing import Final, Sequence, TypeGuard

from .errors import (
    InvalidConfigKeyError,
    InvalidConfigSectionError,
    InvalidConfigValueError,
    InvalidConfigValueTypeError,
)


def parse_config_key(key: str | Sequence[str]) -> list[str]:
    if isinstance(key, str):
        return key.split(".")
    return list(key)


SECTION_INSTALLATION: Final = "installation"
KEY_INSTALLATION_EXTERNALLY_MANAGED: Final = "externally_managed"

SECTION_PACKAGES: Final = "packages"
KEY_PACKAGES_PRERELEASES: Final = "prereleases"

SECTION_REPO: Final = "repo"
KEY_REPO_BRANCH: Final = "branch"
KEY_REPO_LOCAL: Final = "local"
KEY_REPO_REMOTE: Final = "remote"

SECTION_REPOS: Final = "repos"
KEY_REPOS_ID: Final = "id"
KEY_REPOS_NAME: Final = "name"
KEY_REPOS_REMOTE: Final = "remote"
KEY_REPOS_BRANCH: Final = "branch"
KEY_REPOS_LOCAL: Final = "local"
KEY_REPOS_PRIORITY: Final = "priority"
KEY_REPOS_ACTIVE: Final = "active"

SECTION_TELEMETRY: Final = "telemetry"
KEY_TELEMETRY_MODE: Final = "mode"
KEY_TELEMETRY_PM_TELEMETRY_URL: Final = "pm_telemetry_url"
KEY_TELEMETRY_UPLOAD_CONSENT: Final = "upload_consent"


def validate_section(section: str) -> None:
    if section not in (
        SECTION_INSTALLATION,
        SECTION_PACKAGES,
        SECTION_REPO,
        SECTION_REPOS,
        SECTION_TELEMETRY,
    ):
        raise InvalidConfigSectionError(section)


def get_expected_type_for_config_key(key: str | Sequence[str]) -> type | Sequence[type]:
    parsed_key = parse_config_key(key)
    if len(parsed_key) != 2:
        # for now there's no nested config option
        raise InvalidConfigKeyError(key)

    section, sel = parsed_key
    if section == SECTION_INSTALLATION:
        return _get_expected_type_for_section_installation(sel)
    elif section == SECTION_PACKAGES:
        return _get_expected_type_for_section_packages(sel)
    elif section == SECTION_REPO:
        return _get_expected_type_for_section_repo(sel)
    elif section == SECTION_TELEMETRY:
        return _get_expected_type_for_section_telemetry(sel)
    else:
        raise InvalidConfigKeyError(key)


def _get_expected_type_for_section_installation(sel: str) -> type:
    if sel == KEY_INSTALLATION_EXTERNALLY_MANAGED:
        return bool
    else:
        raise InvalidConfigKeyError(sel)


def _get_expected_type_for_section_packages(sel: str) -> type:
    if sel == KEY_PACKAGES_PRERELEASES:
        return bool
    else:
        raise InvalidConfigKeyError(sel)


def _get_expected_type_for_section_repo(sel: str) -> type:
    if sel == KEY_REPO_BRANCH:
        return str
    elif sel == KEY_REPO_LOCAL:
        return str
    elif sel == KEY_REPO_REMOTE:
        return str
    else:
        raise InvalidConfigKeyError(sel)


def _get_expected_type_for_section_telemetry(sel: str) -> type | tuple[type, ...]:
    if sel == KEY_TELEMETRY_MODE:
        return str
    elif sel == KEY_TELEMETRY_PM_TELEMETRY_URL:
        return str
    elif sel == KEY_TELEMETRY_UPLOAD_CONSENT:
        return (type(None), datetime.datetime)
    else:
        raise InvalidConfigKeyError(sel)


def _is_all_str(obj: object) -> TypeGuard[Sequence[str]]:
    if not isinstance(obj, Sequence):
        return False
    return all(isinstance(i, str) for i in obj)


def _is_all_type(obj: object) -> TypeGuard[Sequence[type]]:
    if not isinstance(obj, Sequence):
        return False
    return all(isinstance(i, type) for i in obj)


def ensure_valid_config_kv(
    key: str | Sequence[str],
    check_val: bool = False,
    val: object | None = None,
) -> None:
    parsed_key = parse_config_key(key)
    if len(parsed_key) != 2:
        # for now there's no nested config option
        raise InvalidConfigKeyError(key)

    expected_types = get_expected_type_for_config_key(parsed_key)
    # validity of config key is already checked by get_expected_type_for_config_key
    ensure_value_type(key, check_val, val, expected_types)

    if not check_val:
        return

    section, sel = parsed_key
    if section == SECTION_TELEMETRY:
        return _extra_validate_section_telemetry_kv(key, sel, val)


def ensure_value_type(
    key: str | Sequence[str],
    check_val: bool,
    val: object | None,
    expected: type | Sequence[type],
) -> None:
    if not check_val:
        return

    expected_types: tuple[type, ...]
    if isinstance(expected, type):
        expected_types = (expected,)
    else:
        expected_types = tuple(expected)

    for ty in expected_types:
        if isinstance(val, ty):
            return

    raise InvalidConfigValueTypeError(key, val, expected)


def _extra_validate_section_telemetry_kv(
    key: str | Sequence[str],
    sel: str,
    val: object | None,
) -> None:
    if sel == KEY_TELEMETRY_MODE:
        # value type is already ensured earlier
        if val not in ("local", "off", "on"):
            raise InvalidConfigValueError(key, val, str)


def encode_value(v: object) -> str:
    """Encodes the given config value into a string representation suitable for
    display or storage into TOML config files."""

    if v is None:
        from ..utils.toml import NoneValue

        raise NoneValue()

    if isinstance(v, bool):
        return "true" if v else "false"
    elif isinstance(v, int):
        return str(v)
    elif isinstance(v, str):
        return v
    elif isinstance(v, datetime.datetime):
        if v.tzinfo is None:
            raise ValueError("only timezone-aware datetimes are supported for safety")
        s = v.isoformat()
        if s.endswith("+00:00"):
            # use the shorter 'Z' suffix for UTC
            return f"{s[:-6]}Z"
        return s
    else:
        raise NotImplementedError(f"invalid type for config value: {type(v)}")


def decode_value(
    key: str | Sequence[str],
    val: str,
) -> object:
    """Decodes the given string representation of a config value into a Python
    value, directed by type information implied by the config key."""

    if isinstance(key, type):
        return _decode_single_type_value(None, val, key)
    elif _is_all_type(key):
        return _decode_typed_value(None, val, key)

    assert isinstance(key, str) or _is_all_str(key)
    expected_types = get_expected_type_for_config_key(key)

    if isinstance(expected_types, type):
        return _decode_single_type_value(key, val, expected_types)
    return _decode_typed_value(key, val, expected_types)


def _decode_typed_value(
    key: str | Sequence[str] | None,
    val: str,
    expected_types: Sequence[type],
) -> object:
    for ty in expected_types:
        try:
            return _decode_single_type_value(key, val, ty)
        except (RuntimeError, TypeError, ValueError):
            continue
    raise InvalidConfigValueError(key, val, expected_types)


def _decode_single_type_value(
    key: str | Sequence[str] | None,
    val: str,
    expected_type: type,
) -> object:
    if expected_type is bool:
        if val in ("true", "yes", "1"):
            return True
        elif val in ("false", "no", "0"):
            return False
        else:
            raise InvalidConfigValueError(key, val, expected_type)
    elif expected_type is int:
        return int(val, 10)
    elif expected_type is str:
        return val
    elif expected_type is datetime.datetime:
        # datetime.fromisoformat() has supported the 'Z' suffix since Python 3.11
        v = datetime.datetime.fromisoformat(val)
        return v.astimezone() if v.tzinfo is None else v
    else:
        raise NotImplementedError(f"unhandled type for config value: {expected_type}")



import pathlib
import sys

from ruyi.log import RuyiLogger
from ruyi.ruyipkg import unpack
from ruyi.ruyipkg.unpack_method import UnpackMethod, determine_unpack_method

from ..fixtures import RuyiFileFixtureFactory


def test_unpack_deb(
    ruyi_file: RuyiFileFixtureFactory,
    ruyi_logger: RuyiLogger,
    tmp_path: pathlib.Path,
) -> None:
    with ruyi_file.path("cpp-for-host_14-20240120-6_riscv64.deb") as p:
        assert determine_unpack_method(str(p)) == UnpackMethod.DEB
        unpack.do_unpack(
            ruyi_logger,
            str(p),
            str(tmp_path),
            0,
            UnpackMethod.DEB,
            None,
        )
        check = tmp_path / "usr" / "share" / "doc" / "cpp-for-host"
        if sys.version_info >= (3, 12):
            assert check.exists(follow_symlinks=False)
        else:
            # Python 3.11 lacks pathlib.Path.exists(follow_symlinks)
            #
            # we know that this path is going to be a symlink so simply
            # ensuring it's existent is enough; asserting that it is dangling
            # risks breaking CI on systems where the target actually exists
            assert check.lstat() is not None
        assert check.is_symlink()
        assert str(check.readlink()) == "cpp-riscv64-linux-gnu"



import dataclasses
import enum
import pathlib
import tomllib
from typing import Iterable, Literal, NotRequired, Sequence

import tomlkit
from tomlkit.exceptions import ParseError

from ..utils.porcelain import PorcelainEntity, PorcelainEntityType
from .canonical_dump import dumps_canonical_package_manifest_toml
from .host import get_native_host
from .list_filter import ListFilter, ListFilterOp, ListFilterOpKind
from .pkg_manifest import PackageManifest
from .repo import RepoConfig

try:
    from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
except ModuleNotFoundError:
    from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]


CheckName = Literal["format", "parse"]
CheckSet = frozenset[CheckName]

DEFAULT_CHECKS: CheckSet = frozenset(("format", "parse"))

CHECK_FORMAT: CheckName = "format"
CHECK_PARSE: CheckName = "parse"


class CheckSeverity(enum.StrEnum):
    ERROR = "error"
    WARNING = "warning"


class CheckUsageError(ValueError):
    pass


class CheckDiagnosticPorcelain(PorcelainEntity, total=False):
    severity: str
    code: str
    check: str
    path: str
    message: str
    line: NotRequired[int]
    column: NotRequired[int]
    hint: NotRequired[str]


@dataclasses.dataclass(frozen=True)
class CheckDiagnostic:
    severity: CheckSeverity
    code: str
    check: str
    path: pathlib.Path
    message: str
    line: int | None = None
    column: int | None = None
    hint: str | None = None

    def to_porcelain(self) -> CheckDiagnosticPorcelain:
        result: CheckDiagnosticPorcelain = {
            "ty": PorcelainEntityType.CheckDiagnosticV1,
            "severity": self.severity.value,
            "code": self.code,
            "check": self.check,
            "path": str(self.path),
            "message": self.message,
        }
        if self.line is not None:
            result["line"] = self.line
        if self.column is not None:
            result["column"] = self.column
        if self.hint is not None:
            result["hint"] = self.hint
        return result


@dataclasses.dataclass(frozen=True)
class ManifestRepoContext:
    repo_root: pathlib.Path
    manifest_root_name: str
    category: str
    name: str
    version: str


def normalize_checks(raw_checks: Iterable[str] | None) -> CheckSet:
    if raw_checks is None:
        return DEFAULT_CHECKS

    checks: set[CheckName] = set()
    for raw_check in raw_checks:
        match raw_check:
            case "format":
                checks.add(CHECK_FORMAT)
            case "parse":
                checks.add(CHECK_PARSE)
            case _:
                raise CheckUsageError(f"unsupported check: {raw_check}")
    return frozenset(checks) if checks else DEFAULT_CHECKS


def infer_manifest_repo_context(path: pathlib.Path) -> ManifestRepoContext | None:
    if len(path.parents) < 4:
        return None

    manifest_root = path.parents[2]
    if manifest_root.name not in ("packages", "manifests"):
        return None

    return ManifestRepoContext(
        repo_root=path.parents[3],
        manifest_root_name=manifest_root.name,
        category=path.parents[1].name,
        name=path.parent.name,
        version=path.stem,
    )


def parse_package_selector_args(tokens: Sequence[str]) -> ListFilter:
    if not tokens:
        raise CheckUsageError("--only-packages requires at least one package selector")

    result = ListFilter()
    idx = 0
    unary_options = {
        "--category-contains": ListFilterOpKind.CATEGORY_CONTAINS,
        "--category-is": ListFilterOpKind.CATEGORY_IS,
        "--name-contains": ListFilterOpKind.NAME_CONTAINS,
    }
    unsupported_options = {"--is-installed", "--related-to-entity"}

    while idx < len(tokens):
        token = tokens[idx]
        option, sep, inline_value = token.partition("=")

        if option == "--all":
            if sep:
                raise CheckUsageError("--all does not take an argument")
            result.append(ListFilterOp(ListFilterOpKind.ALL, ""))
            idx += 1
            continue

        if option in unsupported_options:
            raise CheckUsageError(
                f"package selector {option} is not supported by admin check"
            )

        op_kind = unary_options.get(option)
        if op_kind is None:
            raise CheckUsageError(f"unsupported package selector: {token}")

        if sep:
            value = inline_value
        else:
            idx += 1
            if idx >= len(tokens):
                raise CheckUsageError(f"package selector {option} requires an argument")
            value = tokens[idx]
        result.append(ListFilterOp(op_kind, value))
        idx += 1

    return result


def check_manifest_file(
    path: pathlib.Path,
    checks: CheckSet = DEFAULT_CHECKS,
) -> list[CheckDiagnostic]:
    diagnostics: list[CheckDiagnostic] = []

    text = _read_manifest_text(path, diagnostics)
    if text is None:
        return diagnostics

    manifest = _parse_manifest_text(path, text, diagnostics)
    if manifest is None:
        return diagnostics

    if CHECK_PARSE in checks:
        diagnostics.extend(_check_manifest_parse_surface(path, manifest))
        if diagnostics:
            return diagnostics

    if CHECK_FORMAT in checks:
        diagnostics.extend(_check_manifest_format(path, text, manifest))

    return diagnostics


def check_repo(
    repo_root: pathlib.Path,
    checks: CheckSet = DEFAULT_CHECKS,
    package_selector: ListFilter | None = None,
) -> list[CheckDiagnostic]:
    diagnostics = check_repo_config(repo_root)

    manifest_root = _find_manifest_root(repo_root)
    if manifest_root is None:
        return diagnostics

    for path in sorted(p for p in manifest_root.rglob("*") if p.is_file()):
        rel_path = path.relative_to(manifest_root)
        parts = rel_path.parts
        identity = _package_identity_from_manifest_relpath(parts)

        if (
            package_selector is not None
            and identity is not None
            and not _package_selector_matches(package_selector, *identity)
        ):
            continue

        path_diagnostic = _check_repo_manifest_path(path, rel_path, parts)
        if path_diagnostic is not None:
            diagnostics.append(path_diagnostic)
            continue

        diagnostics.extend(check_manifest_file(path, checks))

    return diagnostics


def check_repo_config(repo_root: pathlib.Path) -> list[CheckDiagnostic]:
    path = repo_root / "config.toml"
    try:
        with open(path, "rb") as fp:
            obj = tomllib.load(fp)
    except FileNotFoundError:
        return [
            _diagnostic(
                "RYC0005",
                CHECK_PARSE,
                path,
                "config.toml is missing",
            )
        ]
    except tomllib.TOMLDecodeError as exc:
        return [
            _diagnostic(
                "RYC0005",
                CHECK_PARSE,
                path,
                str(exc),
                line=getattr(exc, "lineno", None),
                column=getattr(exc, "colno", None),
            )
        ]
    except (OSError, UnicodeDecodeError) as exc:
        return [
            _diagnostic(
                "RYC0005",
                CHECK_PARSE,
                path,
                str(exc),
            )
        ]

    try:
        RepoConfig.from_object(obj)
    except (KeyError, TypeError, ValueError, RuntimeError) as exc:
        return [
            _diagnostic(
                "RYC0005",
                CHECK_PARSE,
                path,
                _format_manifest_error(exc),
            )
        ]
    return []


def _read_manifest_text(
    path: pathlib.Path,
    diagnostics: list[CheckDiagnostic],
) -> str | None:
    try:
        return path.read_text(encoding="utf-8")
    except (OSError, UnicodeDecodeError) as exc:
        diagnostics.append(
            _diagnostic(
                "RYC0002",
                CHECK_PARSE,
                path,
                str(exc),
            )
        )
        return None


def _parse_manifest_text(
    path: pathlib.Path,
    text: str,
    diagnostics: list[CheckDiagnostic],
) -> PackageManifest | None:
    try:
        doc = tomlkit.loads(text)
    except ParseError as exc:
        diagnostics.append(
            _diagnostic(
                "RYC0002",
                CHECK_PARSE,
                path,
                str(exc),
                line=getattr(exc, "line", None),
                column=getattr(exc, "col", None),
            )
        )
        return None
    except ValueError as exc:
        diagnostics.append(
            _diagnostic(
                "RYC0002",
                CHECK_PARSE,
                path,
                str(exc),
            )
        )
        return None

    try:
        return PackageManifest(doc)
    except (KeyError, TypeError, ValueError, RuntimeError) as exc:
        diagnostics.append(
            _diagnostic(
                "RYC0003",
                CHECK_PARSE,
                path,
                _format_manifest_error(exc),
            )
        )
        return None


def _check_manifest_parse_surface(
    path: pathlib.Path,
    manifest: PackageManifest,
) -> list[CheckDiagnostic]:
    try:
        _touch_manifest_parse_surface(manifest)
    except (KeyError, TypeError, ValueError, RuntimeError) as exc:
        return [
            _diagnostic(
                "RYC0003",
                CHECK_PARSE,
                path,
                _format_manifest_error(exc),
            )
        ]
    return []


def _check_manifest_format(
    path: pathlib.Path,
    text: str,
    manifest: PackageManifest,
) -> list[CheckDiagnostic]:
    try:
        canonical_text = dumps_canonical_package_manifest_toml(manifest)
    except (KeyError, TypeError, ValueError, RuntimeError) as exc:
        return [
            _diagnostic(
                "RYC0003",
                CHECK_FORMAT,
                path,
                _format_manifest_error(exc),
            )
        ]

    if text == canonical_text:
        return []

    return [
        _diagnostic(
            "RYC0001",
            CHECK_FORMAT,
            path,
            "manifest is not canonical",
            hint=f"run: ruyi admin format-manifest {path}",
        )
    ]


def _touch_manifest_parse_surface(manifest: PackageManifest) -> None:
    manifest.to_raw()
    manifest.raw_doc
    manifest.slug
    manifest.kind
    manifest.desc
    manifest.doc_uri
    manifest.vendor_name
    manifest.upstream_version

    service_level = manifest.service_level
    service_level.level
    service_level.has_known_issues
    list(service_level.known_issues)

    for distfile in manifest.distfiles.values():
        distfile.name
        distfile.urls
        distfile.size
        distfile.checksums
        distfile.prefixes_to_unpack
        distfile.strip_components
        distfile.unpack_method
        distfile.fetch_restriction
        distfile.get_checksum("sha256")
        distfile.is_restricted("fetch")
        distfile.is_restricted("mirror")

    if binary_metadata := manifest.binary_metadata:
        binary_metadata.data
        binary_metadata.is_available_for_current_host
        for host in binary_metadata.data:
            binary_metadata.get_distfile_names_for_host(host)
            binary_metadata.get_commands_for_host(host)

    if blob_metadata := manifest.blob_metadata:
        blob_metadata.get_distfile_names()

    native_host = get_native_host()
    if source_metadata := manifest.source_metadata:
        source_metadata.get_distfile_names_for_host(native_host)

    if toolchain_metadata := manifest.toolchain_metadata:
        toolchain_metadata.target
        toolchain_metadata.target_arch
        toolchain_metadata.quirks
        toolchain_metadata.has_quirk("default")
        toolchain_metadata.satisfies_quirk_set(set())
        list(toolchain_metadata.components)
        toolchain_metadata.get_component_version("gcc")
        toolchain_metadata.has_binutils
        toolchain_metadata.has_clang
        toolchain_metadata.has_gcc
        toolchain_metadata.has_llvm
        toolchain_metadata.included_sysroot

    if emulator_metadata := manifest.emulator_metadata:
        emulator_metadata.quirks
        for program in emulator_metadata.programs:
            program.relative_path
            program.flavor
            program.supported_arches
            program.binfmt_misc
            program.is_qemu
        list(emulator_metadata.list_for_arch(native_host.arch))

    if provisionable_metadata := manifest.provisionable_metadata:
        provisionable_metadata.partition_map
        provisionable_metadata.strategy


def _find_manifest_root(repo_root: pathlib.Path) -> pathlib.Path | None:
    for name in ("packages", "manifests"):
        manifest_root = repo_root / name
        if manifest_root.is_dir():
            return manifest_root
    return None


def _package_identity_from_manifest_relpath(
    parts: tuple[str, ...],
) -> tuple[str, str] | None:
    if len(parts) < 2:
        return None
    return parts[0], parts[1]


def _package_selector_matches(
    package_selector: ListFilter,
    category: str,
    pkg_name: str,
) -> bool:
    for op in package_selector.ops:
        match op.op:
            case ListFilterOpKind.ALL:
                continue
            case ListFilterOpKind.CATEGORY_CONTAINS:
                if op.arg not in category:
                    return False
            case ListFilterOpKind.CATEGORY_IS:
                if op.arg != category:
                    return False
            case ListFilterOpKind.NAME_CONTAINS:
                if op.arg not in pkg_name:
                    return False
            case ListFilterOpKind.RELATED_TO_ENTITY | ListFilterOpKind.IS_INSTALLED:
                raise CheckUsageError(
                    "stateful package selectors are not supported by admin check"
                )
            case _:
                raise CheckUsageError("unknown package selector")
    return True


def _check_repo_manifest_path(
    path: pathlib.Path,
    rel_path: pathlib.Path,
    parts: tuple[str, ...],
) -> CheckDiagnostic | None:
    if len(parts) != 3:
        return _diagnostic(
            "RYC0004",
            CHECK_PARSE,
            path,
            f"manifest path must be <category>/<name>/<version>.toml, got {rel_path}",
        )

    if path.suffix.lower() != ".toml":
        return _diagnostic(
            "RYC0004",
            CHECK_PARSE,
            path,
            "manifest file must use the .toml extension",
        )

    version = path.stem
    try:
        Version.parse(version)
    except ValueError as exc:
        return _diagnostic(
            "RYC0004",
            CHECK_PARSE,
            path,
            f"manifest filename is not a valid semantic version: {version}: {exc}",
        )
    return None


def _diagnostic(
    code: str,
    check: CheckName,
    path: pathlib.Path,
    message: str,
    *,
    line: int | None = None,
    column: int | None = None,
    hint: str | None = None,
) -> CheckDiagnostic:
    return CheckDiagnostic(
        severity=CheckSeverity.ERROR,
        code=code,
        check=check,
        path=path,
        message=message,
        line=line,
        column=column,
        hint=hint,
    )


def _format_manifest_error(exc: BaseException) -> str:
    if isinstance(exc, KeyError):
        return f"missing package manifest field: {exc!s}"
    message = str(exc)
    return message if message else exc.__class__.__name__



"""Tests for RuyiBuildRecipeAPI (B5)."""

import pathlib
from typing import Any

import pytest

from ruyi.log import RuyiLogger
from ruyi.pluginhost.api import RuyiHostAPI
from ruyi.pluginhost.build_api import RuyiBuildRecipeAPI, ScheduledBuild
from ruyi.pluginhost.ctx import PluginHostContext


def _make_recipe_phctx(
    tmp_path: pathlib.Path,
    ruyi_logger: RuyiLogger,
) -> PluginHostContext[Any, Any]:
    plugin_root = tmp_path / "plugins"
    plugin_root.mkdir()
    recipe_root = tmp_path / "recipes_proj"
    recipe_root.mkdir()
    return PluginHostContext.new(
        ruyi_logger,
        plugin_root,
        recipe_project_root=recipe_root,
    )


def _make_plain_phctx(
    tmp_path: pathlib.Path,
    ruyi_logger: RuyiLogger,
) -> PluginHostContext[Any, Any]:
    plugin_root = tmp_path / "plugins"
    plugin_root.mkdir()
    return PluginHostContext.new(ruyi_logger, plugin_root)


def test_schedule_build_registers_callable(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"

    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    def build_one(ctx: object) -> None:
        pass

    api.schedule_build(build_one)

    registry = phctx.scheduled_builds_for(recipe_file)
    assert len(registry) == 1
    sb = registry[0]
    assert isinstance(sb, ScheduledBuild)
    assert sb.name == "build_one"
    assert sb.fn is build_one
    assert sb.recipe_file == recipe_file


def test_schedule_build_explicit_name(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    api.schedule_build(lambda ctx: None, name="custom")

    registry = phctx.scheduled_builds_for(recipe_file)
    assert [sb.name for sb in registry] == ["custom"]


def test_schedule_build_rejects_non_callable(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    with pytest.raises(RuntimeError, match="expected a callable"):
        api.schedule_build("not a function")


def test_schedule_build_accepts_lambda(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    api.schedule_build(lambda ctx: None)
    assert phctx.scheduled_builds_for(recipe_file)[0].name == "<lambda>"


def test_schedule_build_rejects_callable_without_derivable_name(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    class CallableWithoutName:
        def __call__(self, ctx: Any) -> None:
            pass

    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    with pytest.raises(RuntimeError, match="could not derive a name"):
        api.schedule_build(CallableWithoutName())


def test_schedule_build_rejects_duplicate_name(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    api = RuyiBuildRecipeAPI(phctx, recipe_file)

    api.schedule_build(lambda ctx: None, name="dup")
    with pytest.raises(RuntimeError, match="duplicate build name"):
        api.schedule_build(lambda ctx: None, name="dup")


def test_recipe_phctx_has_build_recipe_capability(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    assert "build-recipe-v1" in phctx.capabilities
    # Recipes must go through ctx.subprocess; raw subprocess is denied.
    assert "call-subprocess-v1" not in phctx.capabilities


def test_plain_phctx_does_not_have_build_recipe_capability(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_plain_phctx(tmp_path, ruyi_logger)
    assert "build-recipe-v1" not in phctx.capabilities
    assert "call-subprocess-v1" in phctx.capabilities


def test_build_namespace_gated_on_capability(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_plain_phctx(tmp_path, ruyi_logger)
    host_api = RuyiHostAPI(
        phctx,
        this_file=tmp_path / "unused.star",
        this_plugin_dir=tmp_path,
        allow_host_fs_access=False,
    )
    with pytest.raises(
        RuntimeError, match="only available when loading a build recipe"
    ):
        _ = host_api.build


def test_build_namespace_available_in_recipe_context(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    host_api = RuyiHostAPI(
        phctx,
        this_file=recipe_file,
        this_plugin_dir=tmp_path / "recipes_proj",
        allow_host_fs_access=False,
    )
    build = host_api.build
    assert isinstance(build, RuyiBuildRecipeAPI)


def test_call_subprocess_rejected_in_recipe_context(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    phctx = _make_recipe_phctx(tmp_path, ruyi_logger)
    recipe_file = tmp_path / "recipes_proj" / "pkg.star"
    host_api = RuyiHostAPI(
        phctx,
        this_file=recipe_file,
        this_plugin_dir=tmp_path / "recipes_proj",
        allow_host_fs_access=False,
    )
    with pytest.raises(RuntimeError, match="call_subprocess_argv is not available"):
        host_api.call_subprocess_argv(["/bin/true"])



import abc
import re
from typing import Final, Iterator, Literal, Tuple

import semver

from .pkg_manifest import BoundPackageManifest, is_prerelease
from .protocols import ProvidesPackageManifests


AtomKind = Literal["name"] | Literal["expr"] | Literal["slug"]

RE_ATOM_EXPR: Final = re.compile(r"^([^:(]+)\((.+)\)$")
RE_ATOM_NAME: Final = re.compile(r"^[^:()]+$")


class Atom(abc.ABC):
    def __init__(self, s: str, kind: AtomKind) -> None:
        self._s = s
        self._kind: AtomKind = kind

    @property
    def input_str(self) -> str:
        return self._s

    @property
    def kind(self) -> AtomKind:
        return self._kind

    @classmethod
    def parse(cls, s: str) -> "SlugAtom | NameAtom | ExprAtom":
        if s.startswith("slug:"):
            return SlugAtom(s)

        if s.startswith("name:"):
            return NameAtom(s, s[5:])  # strip the "name:" prefix

        if match := RE_ATOM_EXPR.match(s):
            return ExprAtom(s, match[1], match[2])

        # fallback
        if match := RE_ATOM_NAME.match(s):
            return NameAtom(s, s)

        raise ValueError(f"invalid atom: '{s}'")

    @abc.abstractmethod
    def match_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> BoundPackageManifest | None:
        raise NotImplementedError

    @abc.abstractmethod
    def iter_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> Iterator[BoundPackageManifest]:
        raise NotImplementedError


def split_category(name: str) -> Tuple[str | None, str]:
    fragments = name.split("/", 1)
    if len(fragments) == 2:
        return (fragments[0], fragments[1])
    return (None, name)


class NameAtom(Atom):
    def __init__(self, s: str, name: str) -> None:
        super().__init__(s, "name")

        self.category, self.name = split_category(name)

    def match_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> BoundPackageManifest | None:
        # return the latest version of the package named self.name in the given repo
        try:
            return repo.get_pkg_latest_ver(
                self.name,
                self.category,
                include_prerelease_vers,
            )
        except KeyError:
            return None

    def iter_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> Iterator[BoundPackageManifest]:
        # return all versions of the package named self.name in the given repo
        for pm in repo.iter_pkg_vers(self.name, self.category):
            if not is_prerelease(pm.semver) or include_prerelease_vers:
                yield pm


def fix_version_matcher_for_semver2(match_expr: str) -> str:
    # equivalent of https://github.com/python-semver/python-semver/pull/362
    # for semver 2.x
    if match_expr and match_expr[0] in "0123456789":
        return f"=={match_expr}"
    return match_expr


class ExprAtom(Atom):
    def __init__(self, s: str, name: str, expr: str) -> None:
        super().__init__(s, "expr")
        self.exprs = expr.split(",")

        if semver.__version__ < "3":
            self.exprs = list(map(fix_version_matcher_for_semver2, self.exprs))

        self.category, self.name = split_category(name)

    def _is_pm_matching_my_exprs(self, pm: BoundPackageManifest) -> bool:
        for e in self.exprs:
            if not pm.semver.match(e):
                return False
        return True

    def match_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> BoundPackageManifest | None:
        matching_pms = {
            pm.ver: pm
            for pm in repo.iter_pkg_vers(self.name, self.category)
            if self._is_pm_matching_my_exprs(pm)
        }
        if not matching_pms:
            return None

        semvers = [pm.semver for pm in matching_pms.values()]
        if not include_prerelease_vers:
            semvers = [sv for sv in semvers if not is_prerelease(sv)]
        if not semvers:
            return None
        latest_ver = max(semvers)
        return matching_pms[str(latest_ver)]

    def iter_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> Iterator[BoundPackageManifest]:
        for pm in repo.iter_pkg_vers(self.name, self.category):
            if self._is_pm_matching_my_exprs(pm):
                if not is_prerelease(pm.semver) or include_prerelease_vers:
                    yield pm


class SlugAtom(Atom):
    def __init__(self, s: str) -> None:
        super().__init__(s, "slug")
        self.slug = s[5:]  # strip the "slug:" prefix

    def match_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> BoundPackageManifest | None:
        pm = repo.get_pkg_by_slug(self.slug)
        if pm and is_prerelease(pm.semver):
            return pm if include_prerelease_vers else None
        return pm

    def iter_in_repo(
        self,
        repo: ProvidesPackageManifests,
        include_prerelease_vers: bool,
    ) -> Iterator[BoundPackageManifest]:
        pm = repo.get_pkg_by_slug(self.slug)
        if pm is None:
            return None

        if is_prerelease(pm.semver) and include_prerelease_vers:
            yield pm



import argparse
import enum
from typing import Any, Callable, Iterable, NamedTuple, Sequence, TypeVar, TYPE_CHECKING

from ..cli.completion import ArgcompleteAction
from ..utils.global_mode import TRUTHY_ENV_VAR_VALUES

if TYPE_CHECKING:
    from ..config import GlobalConfig
    from .composite_repo import CompositeRepo

_T = TypeVar("_T")


class ListFilterOpKind(enum.Enum):
    UNKNOWN = 0
    CATEGORY_CONTAINS = 1
    CATEGORY_IS = 2
    NAME_CONTAINS = 3
    RELATED_TO_ENTITY = 4
    IS_INSTALLED = 5
    ALL = 6


class ListFilterOp(NamedTuple):
    op: ListFilterOpKind
    arg: str


class ListFilterExecCtx(NamedTuple):
    cfg: "GlobalConfig"
    mr: "CompositeRepo"
    category: str
    pkg_name: str


def _execute_filter_op(op: ListFilterOp, ctx: ListFilterExecCtx) -> bool:
    match op.op:
        case ListFilterOpKind.ALL:
            return True
        case ListFilterOpKind.CATEGORY_CONTAINS:
            return op.arg in ctx.category
        case ListFilterOpKind.CATEGORY_IS:
            return op.arg == ctx.category
        case ListFilterOpKind.NAME_CONTAINS:
            return op.arg in ctx.pkg_name
        case ListFilterOpKind.RELATED_TO_ENTITY:
            es = ctx.mr.entity_store
            return es.is_entity_related_to(
                f"pkg:{ctx.category}/{ctx.pkg_name}",
                op.arg,
                transitive=True,
                unidirectional=False,
            )
        case ListFilterOpKind.IS_INSTALLED:
            asks_for_installed = op.arg.lower() in TRUTHY_ENV_VAR_VALUES

            # We need to check all versions of this package to see if any are installed
            # For now, we'll use a simple heuristic - check if ANY version is installed
            installed_packages = ctx.cfg.ruyipkg_global_state.list_installed_packages()
            is_installed = any(
                pkg.category == ctx.category and pkg.name == ctx.pkg_name
                for pkg in installed_packages
            )
            return not (is_installed ^ asks_for_installed)
        case _:
            return False


class ListFilter:
    def __init__(self) -> None:
        self.ops: list[ListFilterOp] = []

    def __bool__(self) -> bool:
        return len(self.ops) > 0

    def __repr__(self) -> str:
        return f"<ListFilter ops={self.ops!r}>"

    def append(self, op: ListFilterOp) -> None:
        self.ops.append(op)

    def check_pkg_name(
        self,
        cfg: "GlobalConfig",
        mr: "CompositeRepo",
        category: str,
        pkg_name: str,
    ) -> bool:
        ctx = ListFilterExecCtx(cfg, mr, category, pkg_name)
        return all(_execute_filter_op(op, ctx) for op in self.ops)


class ListFilterAction(ArgcompleteAction):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        nargs: int | str | None = None,
        const: _T | None = None,
        default: _T | str | None = None,
        type: Callable[[str], _T] | argparse.FileType | None = None,
        choices: Iterable[_T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        # for now let's just support argument-less and unary filter ops
        if nargs not in (None, 0, 1):
            raise ValueError("nargs not supported")
        if const is not None:
            raise ValueError("const not supported")
        if default is not None:
            raise ValueError("default not supported")
        if type is not None:
            raise ValueError("type not supported")
        if choices is not None:
            raise ValueError("choices not supported")
        if required:
            raise ValueError("required not supported")
        if metavar is None:
            metavar = "STR"

        super().__init__(
            option_strings,
            dest,
            nargs,
            const,
            default,
            type,
            choices,
            required,
            help,
            metavar,
        )

        self.filter_op_kind: ListFilterOpKind
        match option_strings[0].lstrip("-"):
            case "category-contains":
                self.filter_op_kind = ListFilterOpKind.CATEGORY_CONTAINS
            case "category-is":
                self.filter_op_kind = ListFilterOpKind.CATEGORY_IS
            case "name-contains":
                self.filter_op_kind = ListFilterOpKind.NAME_CONTAINS
            case "related-to-entity":
                self.filter_op_kind = ListFilterOpKind.RELATED_TO_ENTITY
            case "is-installed":
                self.filter_op_kind = ListFilterOpKind.IS_INSTALLED
            case "all":
                self.filter_op_kind = ListFilterOpKind.ALL
            case _:
                # should never happen
                self.filter_op_kind = ListFilterOpKind.UNKNOWN

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        dest: ListFilter | None = getattr(namespace, self.dest, None)
        if not dest:
            dest = ListFilter()
            setattr(namespace, self.dest, dest)

        if self.filter_op_kind == ListFilterOpKind.ALL:
            # "--all" takes no argument
            dest.append(ListFilterOp(ListFilterOpKind.ALL, ""))
            return

        val: str
        if isinstance(values, str):
            val = values
        elif isinstance(values, list):
            val = values[0]
        else:
            # should never happen
            # XXX: no easy way to wire to the global logger instance here
            # log.D(f"unexpected values type: {type(values)}")
            val = ""

        dest.append(ListFilterOp(self.filter_op_kind, val))



import os
import pathlib

import pytest

from ruyi.log import RuyiConsoleLogger
from ruyi.ruyipkg.migration import migrate_repo_dir
from tests.fixtures import MockGlobalModeProvider


@pytest.fixture
def mock_logger() -> RuyiConsoleLogger:
    gm = MockGlobalModeProvider()
    return RuyiConsoleLogger(gm)


class TestMigrateRepoDir:
    def test_migrates_legacy_to_new(
        self, tmp_path: pathlib.Path, mock_logger: RuyiConsoleLogger
    ) -> None:
        cache_root = tmp_path / "cache"
        legacy = cache_root / "packages-index"
        legacy.mkdir(parents=True)
        (legacy / "config.toml").write_text("test")

        migrate_repo_dir(str(cache_root), mock_logger)

        new_path = cache_root / "repos" / "ruyisdk"
        assert new_path.is_dir()
        assert (new_path / "config.toml").read_text() == "test"

        # Legacy path should be a symlink to new
        assert legacy.is_symlink()
        assert os.readlink(str(legacy)) == str(new_path)

    def test_noop_if_legacy_missing(
        self, tmp_path: pathlib.Path, mock_logger: RuyiConsoleLogger
    ) -> None:
        cache_root = tmp_path / "cache"
        cache_root.mkdir(parents=True)

        migrate_repo_dir(str(cache_root), mock_logger)

        assert not (cache_root / "repos" / "ruyisdk").exists()

    def test_noop_if_already_migrated(
        self, tmp_path: pathlib.Path, mock_logger: RuyiConsoleLogger
    ) -> None:
        cache_root = tmp_path / "cache"
        new_path = cache_root / "repos" / "ruyisdk"
        new_path.mkdir(parents=True)
        (new_path / "config.toml").write_text("new")

        # Legacy exists as a regular dir too
        legacy = cache_root / "packages-index"
        legacy.mkdir(parents=True)
        (legacy / "config.toml").write_text("old")

        migrate_repo_dir(str(cache_root), mock_logger)

        # Neither should be changed
        assert (new_path / "config.toml").read_text() == "new"
        assert not legacy.is_symlink()
        assert (legacy / "config.toml").read_text() == "old"

    def test_noop_if_legacy_is_symlink(
        self, tmp_path: pathlib.Path, mock_logger: RuyiConsoleLogger
    ) -> None:
        cache_root = tmp_path / "cache"
        new_path = cache_root / "repos" / "ruyisdk"
        new_path.mkdir(parents=True)
        (new_path / "config.toml").write_text("data")

        legacy = cache_root / "packages-index"
        os.symlink(str(new_path), str(legacy))

        migrate_repo_dir(str(cache_root), mock_logger)

        # Symlink still points correctly
        assert legacy.is_symlink()
        assert (legacy / "config.toml").read_text() == "data"



import tomlkit

from ruyi.ruyipkg.canonical_dump import dumps_canonical_package_manifest_toml
from ruyi.ruyipkg.pkg_manifest import PackageManifest

from ..fixtures import RuyiFileFixtureFactory


def test_format_manifest(ruyi_file: RuyiFileFixtureFactory) -> None:
    with ruyi_file.path("ruyipkg_suites", "format_manifest") as fixtures_dir:
        # Find pairs of before/after files
        files = list(fixtures_dir.glob("*.toml"))
        cases = [f.name[:-12] for f in files if f.name.endswith(".before.toml")]

        for case_name in cases:
            # Determine the expected output file name
            before_file = fixtures_dir / f"{case_name}.before.toml"
            after_file = fixtures_dir / f"{case_name}.after.toml"
            assert after_file.exists(), f"Expected file {after_file} does not exist"

            with open(before_file, "rb") as f:
                data = PackageManifest(tomlkit.load(f))

            # Process with the formatter
            result = dumps_canonical_package_manifest_toml(data)

            # Read the expected output
            with open(after_file, "r", encoding="utf-8") as g:
                expected = g.read()

            assert result == expected, (
                f"Formatted output for {before_file.name} doesn't match expected output. "
                f"Check {after_file.name} for the expected formatting result."
            )



"""Recipe-project discovery and marker-file parsing.

A *recipe project* is a directory tree with a ``ruyi-build-recipes.toml``
marker at its root. Build recipes (``.star`` files) anywhere beneath the
root may ``load()`` each other via the ``ruyi-build://`` scheme.

This module owns the on-disk representation: discovery of the project
root given a recipe file, parsing the marker, and a realpath-based join
helper used by loaders and the build runner.
"""

from __future__ import annotations

from dataclasses import dataclass, field
import pathlib
import tomllib
from typing import Any


MARKER_FILENAME = "ruyi-build-recipes.toml"
SUPPORTED_FORMATS = frozenset({"v1"})


class RecipeProjectError(RuntimeError):
    """Raised for any ruyi-build-recipes.toml or project-layout problem."""


@dataclass(frozen=True)
class RecipeProject:
    """A discovered recipe project.

    Attributes:
        root: Absolute, realpath-resolved path to the project root.
        name: Human-readable project name from the marker, or the root
            directory name if unspecified.
        output_dir: Project-relative directory for build outputs; absolute
            path computed as ``root / output_dir``.
        extra_artifact_roots: Absolute allow-list for artifact roots
            outside the project tree. Each entry is realpath-resolved.
    """

    root: pathlib.Path
    name: str
    output_dir: pathlib.Path
    extra_artifact_roots: tuple[pathlib.Path, ...] = field(default_factory=tuple)

    @property
    def marker_path(self) -> pathlib.Path:
        return self.root / MARKER_FILENAME


def discover_recipe_project(recipe_file: pathlib.Path) -> RecipeProject:
    """Find the recipe project containing ``recipe_file``.

    Walks the lexical parents of ``recipe_file`` looking for the marker
    file. The realpath of the recipe file must remain within the realpath
    of the project root (defense against symlink escape). Raises
    :class:`RecipeProjectError` if no marker is found or if the realpath
    check fails.
    """

    if not recipe_file.exists():
        raise RecipeProjectError(f"recipe file not found: {recipe_file}")

    recipe_abs = recipe_file.absolute()
    resolved_recipe = recipe_file.resolve(strict=True)

    for candidate in (recipe_abs.parent, *recipe_abs.parents):
        marker = candidate / MARKER_FILENAME
        if marker.is_file():
            root = candidate.resolve(strict=True)
            if not resolved_recipe.is_relative_to(root):
                raise RecipeProjectError(
                    f"recipe file {recipe_file} escapes its project root {root} "
                    f"after realpath resolution"
                )
            return _parse_marker(root, marker)

    raise RecipeProjectError(
        f"no {MARKER_FILENAME} found in any parent of {recipe_file}"
    )


def _parse_marker(root: pathlib.Path, marker: pathlib.Path) -> RecipeProject:
    try:
        with open(marker, "rb") as f:
            data = tomllib.load(f)
    except tomllib.TOMLDecodeError as e:
        raise RecipeProjectError(f"malformed {MARKER_FILENAME} at {marker}: {e}") from e

    fmt = data.get("format")
    if fmt not in SUPPORTED_FORMATS:
        raise RecipeProjectError(
            f"{marker}: unsupported or missing 'format' (got {fmt!r}; "
            f"expected one of {sorted(SUPPORTED_FORMATS)})"
        )

    project_section: Any = data.get("project", {})
    if not isinstance(project_section, dict):
        raise RecipeProjectError(f"{marker}: [project] must be a table")

    name = project_section.get("name", root.name)
    if not isinstance(name, str) or not name:
        raise RecipeProjectError(f"{marker}: project.name must be a non-empty string")

    output_dir_raw = project_section.get("output_dir", "out")
    if not isinstance(output_dir_raw, str) or not output_dir_raw:
        raise RecipeProjectError(
            f"{marker}: project.output_dir must be a non-empty string"
        )
    output_dir_rel = pathlib.PurePosixPath(output_dir_raw)
    if output_dir_rel.is_absolute():
        raise RecipeProjectError(
            f"{marker}: project.output_dir must be a project-relative path"
        )
    output_dir = (root / output_dir_rel).resolve()
    if not output_dir.is_relative_to(root):
        raise RecipeProjectError(
            f"{marker}: project.output_dir escapes the project root"
        )

    extra_raw = project_section.get("extra_artifact_roots", [])
    if not isinstance(extra_raw, list) or not all(
        isinstance(x, str) for x in extra_raw
    ):
        raise RecipeProjectError(
            f"{marker}: project.extra_artifact_roots must be a list of strings"
        )
    extras: list[pathlib.Path] = []
    for entry in extra_raw:
        p = pathlib.Path(entry)
        if not p.is_absolute():
            raise RecipeProjectError(
                f"{marker}: extra_artifact_roots entries must be absolute paths "
                f"(got {entry!r})"
            )
        extras.append(p.resolve())

    return RecipeProject(
        root=root,
        name=name,
        output_dir=output_dir,
        extra_artifact_roots=tuple(extras),
    )


def safe_join(root: pathlib.Path, rel: str) -> pathlib.Path:
    """Join ``rel`` onto ``root`` and verify the result stays inside ``root``.

    Used when resolving ``ruyi-build://``-scheme paths and ``ctx.repo_path``
    calls. Accepts either forward-slash relative paths or plain relative
    paths; absolute inputs are rejected.
    """

    p = pathlib.PurePosixPath(rel)
    if p.is_absolute():
        raise RecipeProjectError(
            f"safe_join: absolute paths are not allowed (got {rel!r})"
        )
    joined = (root / p).resolve()
    root_resolved = root.resolve()
    if not joined.is_relative_to(root_resolved):
        raise RecipeProjectError(
            f"safe_join: {rel!r} escapes recipe project root {root_resolved}"
        )
    return joined



import pathlib

from ruyi.ruyipkg.check import (
    check_manifest_file,
    check_repo,
    infer_manifest_repo_context,
    parse_package_selector_args,
)


SHA_STUB = "0" * 64


CANONICAL_MANIFEST = f"""format = "v1"

[metadata]
desc = "Test package"
vendor = {{ name = "Test Vendor", eula = "" }}

[[distfiles]]
name = "src.tar.zst"
size = 0

[distfiles.checksums]
sha256 = "{SHA_STUB}"

[source]
distfiles = ["src.tar.zst"]
"""


NON_CANONICAL_MANIFEST = f"""format = "v1"

[[distfiles]]
size = 0
name = "src.tar.zst"

[distfiles.checksums]
sha256 = "{SHA_STUB}"

[metadata]
vendor = {{ eula = "", name = "Test Vendor" }}
desc = "Test package"

[source]
distfiles = ["src.tar.zst"]
"""


VALID_REPO_CONFIG = """ruyi-repo = "v1"

[[mirrors]]
id = "ruyi-dist"
urls = ["https://example.invalid/dist/"]
"""


def _write_manifest(
    root: pathlib.Path,
    category: str,
    name: str,
    version: str,
    text: str = CANONICAL_MANIFEST,
) -> pathlib.Path:
    manifest_dir = root / "packages" / category / name
    manifest_dir.mkdir(parents=True, exist_ok=True)
    manifest_path = manifest_dir / f"{version}.toml"
    manifest_path.write_text(text, encoding="utf-8")
    return manifest_path


def _write_repo_config(root: pathlib.Path, text: str = VALID_REPO_CONFIG) -> None:
    root.mkdir(parents=True, exist_ok=True)
    (root / "config.toml").write_text(text, encoding="utf-8")


def test_clean_canonical_manifest_has_no_diagnostics(
    tmp_path: pathlib.Path,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(CANONICAL_MANIFEST, encoding="utf-8")

    assert check_manifest_file(manifest_path) == []


def test_non_canonical_manifest_reports_format_diagnostic(
    tmp_path: pathlib.Path,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(NON_CANONICAL_MANIFEST, encoding="utf-8")

    diagnostics = check_manifest_file(manifest_path)

    assert [diag.code for diag in diagnostics] == ["RYC0001"]
    assert diagnostics[0].hint == f"run: ruyi admin format-manifest {manifest_path}"


def test_malformed_toml_reports_parse_position(tmp_path: pathlib.Path) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text('format = "v1"\n[metadata\n', encoding="utf-8")

    diagnostics = check_manifest_file(manifest_path)

    assert [diag.code for diag in diagnostics] == ["RYC0002"]
    assert diagnostics[0].line == 2
    assert diagnostics[0].column is not None


def test_unknown_manifest_format_reports_invalid_manifest(
    tmp_path: pathlib.Path,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(
        CANONICAL_MANIFEST.replace('format = "v1"', 'format = "v2"'),
        encoding="utf-8",
    )

    diagnostics = check_manifest_file(manifest_path)

    assert [diag.code for diag in diagnostics] == ["RYC0003"]


def test_missing_required_fields_are_reported_without_traceback(
    tmp_path: pathlib.Path,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(
        f"""format = "v1"

[[distfiles]]
name = "src.tar.zst"
size = 0

[distfiles.checksums]
sha256 = "{SHA_STUB}"
""",
        encoding="utf-8",
    )

    diagnostics = check_manifest_file(manifest_path)
    assert [diag.code for diag in diagnostics] == ["RYC0003"]
    assert "metadata" in diagnostics[0].message


def test_accessor_triggered_malformed_data_is_reported(
    tmp_path: pathlib.Path,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(
        f"""format = "v1"

[metadata]
desc = "Test package"
vendor = {{ name = "Test Vendor", eula = "" }}

[[distfiles]]
size = 0

[distfiles.checksums]
sha256 = "{SHA_STUB}"
""",
        encoding="utf-8",
    )

    diagnostics = check_manifest_file(manifest_path)
    assert [diag.code for diag in diagnostics] == ["RYC0003"]
    assert "name" in diagnostics[0].message


def test_repo_mode_reports_invalid_semver_filenames(
    tmp_path: pathlib.Path,
) -> None:
    _write_repo_config(tmp_path)
    _write_manifest(tmp_path, "source", "sample", "not-semver")

    diagnostics = check_repo(tmp_path)

    assert [diag.code for diag in diagnostics] == ["RYC0004"]


def test_repo_mode_continues_after_bad_files(tmp_path: pathlib.Path) -> None:
    _write_repo_config(tmp_path)
    _write_manifest(tmp_path, "source", "bad-toml", "1.0.0", 'format = "v1"\n[')
    _write_manifest(tmp_path, "source", "bad-version", "not-semver")

    diagnostics = check_repo(tmp_path)

    assert sorted(diag.code for diag in diagnostics) == ["RYC0002", "RYC0004"]


def test_file_mode_infers_repo_context(tmp_path: pathlib.Path) -> None:
    manifest_path = _write_manifest(
        tmp_path,
        "board-image",
        "example-board",
        "1.2.3",
    )

    context = infer_manifest_repo_context(manifest_path)

    assert context is not None
    assert context.repo_root == tmp_path
    assert context.manifest_root_name == "packages"
    assert context.category == "board-image"
    assert context.name == "example-board"
    assert context.version == "1.2.3"


def test_only_packages_limits_package_diagnostics_but_checks_repo_config(
    tmp_path: pathlib.Path,
) -> None:
    _write_repo_config(tmp_path, 'ruyi-repo = "v1"\n')
    _write_manifest(tmp_path, "source", "ignored", "1.0.0", 'format = "v1"\n[')
    board_manifest = _write_manifest(
        tmp_path,
        "board-image",
        "selected",
        "1.0.0",
        NON_CANONICAL_MANIFEST,
    )
    selector = parse_package_selector_args(["--category-is", "board-image"])

    diagnostics = check_repo(tmp_path, package_selector=selector)

    assert [diag.code for diag in diagnostics] == ["RYC0005", "RYC0001"]
    assert diagnostics[1].path == board_manifest



#!/usr/bin/env python3

import os
import pathlib
import sys

import ruyi
from ruyi.i18n import _, ADAPTER
from ruyi.utils.ci import is_running_in_ci
from ruyi.utils.global_mode import (
    EnvGlobalModeProvider,
    ENV_FORCE_ALLOW_ROOT,
    TRUTHY_ENV_VAR_VALUES,
    is_env_var_truthy,
)
from ruyi.utils.node_info import probe_for_container_runtime

# NOTE: no imports that directly or indirectly pull in pygit2 should go here,
# because import of pygit2 will fail if done before ssl_patch. Notably this
# means no GlobalConfig here because it depends on ruyi.ruyipkg.repo.


def _is_running_as_root() -> bool:
    # this is way too simplistic but works on *nix systems which is all we
    # support currently
    if hasattr(os, "getuid"):
        return os.getuid() == 0
    return False


def _is_allowed_to_run_as_root() -> bool:
    if is_env_var_truthy(os.environ, ENV_FORCE_ALLOW_ROOT):
        return True
    if is_running_in_ci(os.environ):
        # CI environments are usually considered to be controlled, and safe
        # for root usage.
        return True
    if probe_for_container_runtime(os.environ) != "unknown":
        # So are container environments.
        return True
    return False


def entrypoint() -> None:
    gm = EnvGlobalModeProvider(os.environ, sys.argv)
    ADAPTER.init_from_env(os.environ)
    ADAPTER.hook()

    # NOTE: import of `ruyi.log` takes ~90ms on my machine, so initialization
    # of logging is deferred as late as possible

    if _is_running_as_root() and not _is_allowed_to_run_as_root():
        from ruyi.log import RuyiConsoleLogger

        logger = RuyiConsoleLogger(gm)

        logger.F(_("refusing to run as super user outside CI without explicit consent"))

        choices = ", ".join(f"'{x}'" for x in TRUTHY_ENV_VAR_VALUES)
        logger.I(
            _(
                "re-run with environment variable [yellow]{env_var}[/] set to one of [yellow]{choices}[/] to signify consent"
            ).format(
                env_var=ENV_FORCE_ALLOW_ROOT,
                choices=choices,
            )
        )
        sys.exit(1)

    if not sys.argv:
        from ruyi.log import RuyiConsoleLogger

        logger = RuyiConsoleLogger(gm)

        logger.F(_("no argv?"))
        sys.exit(1)

    if gm.is_packaged and ruyi.__compiled__.standalone:
        # If we're running from a bundle, our bundled libssl may remember a
        # different path for loading certificates than appropriate for the
        # current system, in which case the pygit2 import will fail. To avoid
        # this we have to patch ssl.get_default_verify_paths with additional
        # logic.
        #
        # this must happen before pygit2 is imported
        from ruyi.utils import ssl_patch

        del ssl_patch

    from ruyi.utils.nuitka import get_nuitka_self_exe, get_argv0

    # note down our own executable path, for identity-checking in mux, if not
    # we're not already Nuitka-compiled
    #
    # we assume the one-file build if Nuitka is detected; sys.argv[0] does NOT
    # work if it's just `ruyi` so we have to check our parent process in that case
    sys.argv[0] = get_argv0()
    if gm.is_packaged:
        self_exe = get_nuitka_self_exe()
    else:
        self_exe = str(pathlib.Path(sys.argv[0]).resolve())
    gm.record_self_exe(sys.argv[0], __file__, self_exe)

    from ruyi.config import GlobalConfig
    from ruyi.cli.main import main
    from ruyi.log import RuyiConsoleLogger

    logger = RuyiConsoleLogger(gm)
    gc = GlobalConfig.load_from_config(gm, logger)
    sys.exit(main(gm, gc, sys.argv))


if __name__ == "__main__":
    entrypoint()



from typing import Callable, TypedDict, TypeGuard, cast

from jinja2 import BaseLoader, Environment, TemplateNotFound

from ..utils.l10n import match_lang_code


RepoMessagesV1Type = TypedDict(
    "RepoMessagesV1Type",
    {
        "ruyi-repo-messages": str,
        # lang_code: message_content
    },
)


def validate_repo_messages_v1(x: object) -> TypeGuard[RepoMessagesV1Type]:
    if not isinstance(x, dict):
        return False
    x = cast(dict[str, object], x)
    if x.get("ruyi-repo-messages", "") != "v1":
        return False
    return True


def group_messages_by_lang_code(decl: RepoMessagesV1Type) -> dict[str, dict[str, str]]:
    obj = cast(dict[str, dict[str, str]], decl)

    result: dict[str, dict[str, str]] = {}
    for msgid, msg_decl in obj.items():
        # skip the file type marker
        if msgid == "ruyi-repo-messages":
            continue

        for lang_code, msg in msg_decl.items():
            if lang_code not in result:
                result[lang_code] = {}
            result[lang_code][msgid] = msg

    return result


class RepoMessageStore:
    def __init__(self, decl: RepoMessagesV1Type) -> None:
        self._msgs_by_lang_code = group_messages_by_lang_code(decl)
        self._cached_envs_by_lang_code: dict[str, Environment] = {}

    @classmethod
    def from_object(cls, obj: object) -> "RepoMessageStore":
        if not validate_repo_messages_v1(obj):
            # TODO: more detail in the error message
            raise RuntimeError("malformed v1 repo messages definition")
        return cls(obj)

    def get_message_template(self, msgid: str, lang_code: str) -> str | None:
        resolved_lang_code = match_lang_code(lang_code, self._msgs_by_lang_code.keys())
        return self._msgs_by_lang_code[resolved_lang_code].get(msgid)

    def get_jinja(self, lang_code: str) -> Environment:
        if lang_code in self._cached_envs_by_lang_code:
            return self._cached_envs_by_lang_code[lang_code]

        env = Environment(
            loader=RepoMessageLoader(self, lang_code),
            autoescape=False,  # we're not producing HTML
            auto_reload=False,  # we're serving static assets
        )
        self._cached_envs_by_lang_code[lang_code] = env
        return env

    def render_message(
        self,
        msgid: str,
        lang_code: str,
        params: dict[str, str],
        add_trailing_newline: bool = False,
    ) -> str:
        env = self.get_jinja(lang_code)
        tmpl = env.get_template(msgid)
        result = tmpl.render(params)
        if add_trailing_newline and not result.endswith("\n"):
            return result + "\n"
        return result


class RepoMessageLoader(BaseLoader):
    def __init__(self, store: RepoMessageStore, lang_code: str) -> None:
        self.store = store
        self.lang_code = lang_code

    def get_source(
        self,
        environment: Environment,
        template: str,
    ) -> tuple[str, (str | None), (Callable[[], bool] | None)]:
        result = self.store.get_message_template(template, self.lang_code)
        if result is None:
            raise TemplateNotFound(template)
        return result, None, None



# Minimal frontmatter support for Markdown, because [python-frontmatter] is
# not packaged in major Linux distributions, complicating packaging work.
#
# Only the YAML frontmatter is supported here, unlike python-frontmatter
# which supports additionally JSON and TOML frontmatter formats.
#
# [python-frontmatter]: https://github.com/eyeseast/python-frontmatter

import re
from typing import Final
import yaml


FRONTMATTER_BOUNDARY_RE: Final = re.compile(r"(?m)^-{3,}\s*$")


class Post:
    def __init__(self, metadata: dict[str, object] | None, content: str) -> None:
        self._md = metadata
        self.content = content

    def get(self, key: str) -> object | None:
        return None if self._md is None else self._md.get(key)


def loads(s: str) -> Post:
    m = FRONTMATTER_BOUNDARY_RE.match(s)
    if m is None:
        return Post(None, s)

    x = FRONTMATTER_BOUNDARY_RE.split(s, 2)
    if len(x) != 3:
        return Post(None, s)

    fm, content = x[1], x[2]

    metadata = yaml.safe_load(fm)
    return Post(metadata, content)



import enum
import re
from typing import Final

from ..i18n import _

RE_TARBALL: Final = re.compile(r"\.tar(?:\.gz|\.bz2|\.lz4|\.xz|\.zst)?$")


class UnpackMethod(enum.StrEnum):
    UNKNOWN = ""
    AUTO = "auto"
    TAR_AUTO = "tar.auto"

    RAW = "raw"
    GZ = "gz"
    BZ2 = "bz2"
    LZ4 = "lz4"
    XZ = "xz"
    ZST = "zst"

    TAR = "tar"
    TAR_GZ = "tar.gz"
    TAR_BZ2 = "tar.bz2"
    TAR_LZ4 = "tar.lz4"
    TAR_XZ = "tar.xz"
    TAR_ZST = "tar.zst"

    ZIP = "zip"
    DEB = "deb"


class UnrecognizedPackFormatError(Exception):
    def __init__(self, filename: str) -> None:
        self.filename = filename

    def __str__(self) -> str:
        return _("don't know how to unpack file {filename}").format(
            filename=self.filename,
        )


def determine_unpack_method(
    filename: str,
) -> UnpackMethod:
    filename_lower = filename.lower()
    if m := RE_TARBALL.search(filename_lower):
        return UnpackMethod(m.group(0)[1:])
    if filename_lower.endswith(".deb"):
        return UnpackMethod.DEB
    if filename_lower.endswith(".zip"):
        return UnpackMethod.ZIP
    if filename_lower.endswith(".gz"):
        # bare gzip file
        return UnpackMethod.GZ
    if filename_lower.endswith(".bz2"):
        # bare bzip2 file
        return UnpackMethod.BZ2
    if filename_lower.endswith(".lz4"):
        # bare lz4 file
        return UnpackMethod.LZ4
    if filename_lower.endswith(".xz"):
        # bare xz file
        return UnpackMethod.XZ
    if filename_lower.endswith(".zst"):
        # bare zstd file
        return UnpackMethod.ZST
    return UnpackMethod.UNKNOWN



from typing import Protocol


class SupportsGetOption(Protocol):
    def get_option(self, key: str) -> object: ...


class SupportsEvalFunction(Protocol):
    def eval_function(
        self,
        function: object,
        *args: object,
        **kwargs: object,
    ) -> object: ...


class SupportsMessageStore(Protocol):
    def get_message_template(self, msgid: str, lang_code: str) -> str | None: ...

    def render_message(
        self,
        msgid: str,
        lang_code: str,
        params: dict[str, str],
        add_trailing_newline: bool = False,
    ) -> str: ...



import base64
import zlib

from .data import RESOURCES, TEMPLATE_NAME_MAP


def _unpack_payload(x: bytes) -> bytes:
    return zlib.decompress(base64.b64decode(x))


_CACHE: dict[str, bytes] = {}


def get_resource_blob(name: str) -> bytes | None:
    if t := RESOURCES.get(name):
        if name not in _CACHE:
            # In our use cases, the program is short-lived and involved resources
            # are small in size, so it is fine to just store the decompressed
            # blobs without eviction.
            _CACHE[name] = _unpack_payload(t)
        return _CACHE[name]
    return None


def get_resource_str(name: str) -> str | None:
    if blob := get_resource_blob(name):
        return blob.decode("utf-8")
    return None


def get_template_str(template_name: str) -> str | None:
    if t := TEMPLATE_NAME_MAP.get(template_name):
        return get_resource_str(t)
    return None



import datetime
from typing import Iterable, TYPE_CHECKING
from unittest.mock import Mock

try:
    from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
except ModuleNotFoundError:
    # semver 2.x
    from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]

from ruyi.ruyipkg.state import BoundInstallationStateStore, PackageInstallationInfo

if TYPE_CHECKING:
    from ruyi.ruyipkg.pkg_manifest import BoundPackageManifest


def test_bound_installation_state_store_empty() -> None:
    """Test BoundInstallationStateStore with no installed packages."""
    # Create a mock RuyipkgGlobalStateStore that returns no installed packages
    mock_rgs = Mock()
    mock_rgs.list_installed_packages.return_value = []

    # Create a mock MetadataRepo
    mock_mr = Mock()

    # Create the BoundInstallationStateStore
    store = BoundInstallationStateStore(mock_rgs, mock_mr)

    # Test that it returns no packages
    assert not list(store.iter_pkg_manifests())
    assert not list(store.iter_pkgs())
    assert not list(store.iter_pkg_vers("nonexistent"))
    assert store.get_pkg_by_slug("nonexistent") is None


def test_bound_installation_state_store_with_installed_packages() -> None:
    """Test BoundInstallationStateStore with some installed packages."""
    # Create mock installed packages
    install_info1 = PackageInstallationInfo(
        repo_id="test-repo",
        category="toolchain",
        name="gcc",
        version="13.1.0",
        host="x86_64-linux-gnu",
        install_path="/test/path1",
        install_time=datetime.datetime.now(),
    )

    install_info2 = PackageInstallationInfo(
        repo_id="test-repo",
        category="toolchain",
        name="gcc",
        version="13.2.0",
        host="x86_64-linux-gnu",
        install_path="/test/path2",
        install_time=datetime.datetime.now(),
    )

    # Create mock manifests
    mock_manifest1 = Mock()
    mock_manifest1.category = "toolchain"
    mock_manifest1.name = "gcc"
    mock_manifest1.ver = "13.1.0"
    mock_manifest1.slug = "gcc-13-1-0"
    mock_manifest1.semver = Version.parse("13.1.0")

    mock_manifest2 = Mock()
    mock_manifest2.category = "toolchain"
    mock_manifest2.name = "gcc"
    mock_manifest2.ver = "13.2.0"
    mock_manifest2.slug = "gcc-13-2-0"
    mock_manifest2.semver = Version.parse("13.2.0")

    # Create a mock RuyipkgGlobalStateStore
    mock_rgs = Mock()
    mock_rgs.list_installed_packages.return_value = [install_info1, install_info2]

    # Create a mock MetadataRepo
    mock_mr = Mock()

    def mock_get_pkg(
        name: str, category: str, ver: str
    ) -> "BoundPackageManifest | None":
        if name == "gcc" and category == "toolchain":
            if ver == "13.1.0":
                return mock_manifest1
            elif ver == "13.2.0":
                return mock_manifest2
        return None

    def mock_iter_pkg_vers(
        name: str, category: str | None = None
    ) -> "Iterable[BoundPackageManifest]":
        if name == "gcc" and category == "toolchain":
            return [mock_manifest1, mock_manifest2]
        return []

    mock_mr.get_pkg.side_effect = mock_get_pkg
    mock_mr.iter_pkg_vers.side_effect = mock_iter_pkg_vers

    # Create the BoundInstallationStateStore
    store = BoundInstallationStateStore(mock_rgs, mock_mr)

    # Test iter_pkg_manifests
    manifests = list(store.iter_pkg_manifests())
    assert len(manifests) == 2
    assert mock_manifest1 in manifests
    assert mock_manifest2 in manifests

    # Test iter_pkgs
    pkgs = list(store.iter_pkgs())
    assert len(pkgs) == 1
    category, name, versions = pkgs[0]
    assert category == "toolchain"
    assert name == "gcc"
    assert len(versions) == 2
    assert "13.1.0" in versions
    assert "13.2.0" in versions

    # Test iter_pkg_vers
    pkg_versions = list(store.iter_pkg_vers("gcc", "toolchain"))
    assert len(pkg_versions) == 2
    assert mock_manifest1 in pkg_versions
    assert mock_manifest2 in pkg_versions

    # Test get_pkg_latest_ver
    latest = store.get_pkg_latest_ver("gcc", "toolchain")
    # Should return the one with the higher version
    assert latest == mock_manifest2

    # Test get_pkg_by_slug
    result = store.get_pkg_by_slug("gcc-13-1-0")
    assert result == mock_manifest1

    # Test get_pkg method
    pkg = store.get_pkg("gcc", "toolchain", "13.1.0")
    assert pkg == mock_manifest1

    pkg = store.get_pkg("gcc", "toolchain", "13.2.0")
    assert pkg == mock_manifest2

    # Test get_pkg with non-existent package
    result = store.get_pkg("nonexistent", "toolchain", "1.0.0")
    assert result is None

    result = store.get_pkg("gcc", "nonexistent", "13.1.0")
    assert result is None

    result = store.get_pkg("gcc", "toolchain", "99.0.0")
    assert result is None



from contextlib import AbstractContextManager
import pathlib
from typing import Any, TYPE_CHECKING

from pygit2 import GitError, Oid
from pygit2.callbacks import RemoteCallbacks
from pygit2.repository import Repository

try:
    from pygit2.remotes import TransferProgress
except ModuleNotFoundError:
    # pygit2 < 1.14.0
    # see https://github.com/libgit2/pygit2/commit/a8b2421bea550292
    #
    # import-untyped: the current pygit2 type stubs were written after the
    # `remote` -> `remotes` rename, so no stubs for it
    from pygit2.remote import TransferProgress  # type: ignore[import-not-found,import-untyped,no-redef,unused-ignore]

# for compatibility with <1.14.0, cannot `from pygit2.enums import MergeAnalysis`
# see https://github.com/libgit2/pygit2/pull/1251
from pygit2 import (
    GIT_MERGE_ANALYSIS_UNBORN,
    GIT_MERGE_ANALYSIS_FASTFORWARD,
    GIT_MERGE_ANALYSIS_UP_TO_DATE,
)

from rich.progress import Progress, TaskID
from rich.text import Text

if TYPE_CHECKING:
    from typing_extensions import Self

from ..i18n import _
from ..log import RuyiLogger


def human_readable_path_of_repo(repo: Repository) -> pathlib.Path:
    """
    Returns a human-readable path of the repository.
    If the repository is a submodule, returns the path to the parent module.
    """
    repo_path = pathlib.Path(repo.path)
    return repo_path.parent if repo_path.name == ".git" else repo_path


class RemoteGitProgressIndicator(
    RemoteCallbacks,
    AbstractContextManager["RemoteGitProgressIndicator"],
):
    def __init__(self) -> None:
        super().__init__()
        self.p = Progress()
        self.task: TaskID | None = None
        self._last_stats: TransferProgress | None = None
        self._task_name: str = ""

    def __enter__(self) -> "Self":
        self.p.__enter__()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, tb: Any) -> None:
        return self.p.__exit__(exc_type, exc_value, tb)

    # Compatibility with pygit2 < 1.8.0.
    def progress(self, string: str) -> None:
        return self.sideband_progress(string)

    def sideband_progress(self, string: str) -> None:
        self.p.console.print("\r", Text(string), sep="", end="")

    def transfer_progress(self, stats: TransferProgress) -> None:
        new_phase = False
        task_name: str = self._task_name
        total: int = 0
        completed: int = 0

        if (
            self._last_stats is None
            or self._last_stats.received_objects != stats.received_objects
        ):
            task_name = _("transferring objects")
            total = stats.total_objects
            completed = stats.received_objects
        elif self._last_stats.indexed_deltas != stats.indexed_deltas:
            task_name = _("processing deltas")
            total = stats.total_deltas
            completed = stats.indexed_deltas
        elif self._last_stats.received_bytes != stats.received_bytes:
            # we don't render the received size at the moment
            pass

        new_phase = self._task_name != task_name
        if new_phase:
            self.task = self.p.add_task(task_name, total=total, completed=completed)
            self._task_name = task_name
        else:
            if self.task is not None:
                self.p.update(self.task, total=total, completed=completed)

        self._last_stats = stats


# based on https://stackoverflow.com/questions/27749418/implementing-pull-with-pygit2
def pull_ff_or_die(
    logger: RuyiLogger,
    repo: Repository,
    remote_name: str,
    remote_url: str,
    branch_name: str,
    *,
    allow_auto_management: bool,
) -> None:
    remote = repo.remotes[remote_name]
    if remote.url != remote_url:
        if not allow_auto_management:
            logger.F(
                _(
                    "URL of remote '[yellow]{remote}[/]' does not match expected URL"
                ).format(remote=remote_name)
            )
            repo_path = human_readable_path_of_repo(repo)
            logger.I(_("repository:          [yellow]{path}[/]").format(path=repo_path))
            logger.I(_("expected remote URL: [yellow]{url}[/]").format(url=remote_url))
            logger.I(_("actual remote URL:   [yellow]{url}[/]").format(url=remote.url))
            logger.I(_("please [bold red]fix the repo settings manually[/]"))
            raise SystemExit(1)

        logger.D(
            f"updating url of remote {remote_name} from {remote.url} to {remote_url}"
        )
        repo.remotes.set_url(remote_name, remote_url)
        # this needs manual refreshing
        remote = repo.remotes[remote_name]

    logger.D("fetching")
    try:
        with RemoteGitProgressIndicator() as pr:
            remote.fetch(callbacks=pr)
    except GitError as e:
        logger.F(
            _("failed to fetch from remote URL {url}: {reason}").format(
                url=remote_url, reason=e
            )
        )
        raise SystemExit(1) from e

    remote_head_ref = repo.lookup_reference(f"refs/remotes/{remote_name}/{branch_name}")
    remote_head: Oid
    if isinstance(remote_head_ref.target, Oid):
        remote_head = remote_head_ref.target
    else:
        assert isinstance(remote_head_ref.target, str)
        remote_head = Oid(hex=remote_head_ref.target)

    merge_analysis, _mp = repo.merge_analysis(remote_head)

    if merge_analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE:
        # nothing to do
        logger.D("repo state already up-to-date")
        return

    if merge_analysis & (GIT_MERGE_ANALYSIS_UNBORN | GIT_MERGE_ANALYSIS_FASTFORWARD):
        # simple fast-forwarding is enough in both cases
        logger.D(f"fast-forwarding repo to {remote_head}")
        tgt = repo.get(remote_head)
        assert tgt is not None
        repo.checkout_tree(tgt)

        logger.D(f"updating branch {branch_name} HEAD")
        local_branch_ref = repo.lookup_reference(f"refs/heads/{branch_name}")
        local_branch_ref.set_target(remote_head)
        repo.head.set_target(remote_head)
        return

    # cannot handle these cases
    logger.F(_("cannot fast-forward repo to newly fetched state"))
    logger.I(_("manual intervention is required to avoid data loss"))
    raise SystemExit(1)



from contextlib import AbstractContextManager
from typing import BinaryIO, TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType

import arpy


class ArpyArchiveWrapper(arpy.Archive, AbstractContextManager["arpy.Archive"]):
    """Compatibility shim for arpy.Archive, for easy interop with both arpy 1.x
    and 2.x."""

    def __init__(
        self,
        filename: str | None = None,
        fileobj: BinaryIO | None = None,
    ) -> None:
        super().__init__(filename=filename, fileobj=fileobj)

    def __enter__(self) -> arpy.Archive:
        if hasattr(super(), "__enter__"):
            # in case we're working with a newer arpy version that has a
            # non-trivial __enter__ implementation
            return super().__enter__()

        # backport of arpy 2.x __enter__ implementation
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: "TracebackType | None",
    ) -> None:
        if hasattr(super(), "__exit__"):
            return super().__exit__(exc_type, exc_value, traceback)

        # backport of arpy 2.x __exit__ implementation
        self.close()

    def infolist(self) -> list[arpy.ArchiveFileHeader]:
        if hasattr(super(), "infolist"):
            return super().infolist()

        # backport of arpy 2.x infolist()
        self.read_all_headers()
        return [
            header
            for header in self.headers
            if header.type
            in (
                arpy.HEADER_BSD,
                arpy.HEADER_NORMAL,
                arpy.HEADER_GNU,
            )
        ]

    def open(self, name: bytes | arpy.ArchiveFileHeader) -> arpy.ArchiveFileData:
        if hasattr(super(), "open"):
            return super().open(name)

        # backport of arpy 2.x open()
        if isinstance(name, bytes):
            ar_file = self.archived_files.get(name)
            if ar_file is None:
                raise KeyError("There is no item named %r in the archive" % (name,))

            return ar_file

        if name not in self.headers:
            raise KeyError("Provided header does not match this archive")

        return arpy.ArchiveFileData(ar_obj=self, header=name)
